#!/usr/bin/perl -w 

sub usage(){
    print STDERR "Usage: $0 -skd <skd file> -c <conf file> [-sim]\n";
    print STDERR " -skd: Schedule file.\n";
    print STDERR " -sim: Simulation mode. Execute recording now.\n";
    print STDERR " -c  : configuration file must be in following format\n";
    print STDERR " -----------------------------\n";
    print STDERR "StationID  G\n";
    print STDERR "Outdir     /mnt/raid/tmp \n";
    print STDERR "NamingRule 2 \n";
    print STDERR "#    1 : XDDDNNNN.dat (default) \n";
    print STDERR "#   -1 : XDDDNNNN.#ch.dat \n";
    print STDERR "#    2 : sidDDDHHMMSSG.dat \n";
    print STDERR "#   -2 : sidDDDHHMMSSG.#ch.dat\n";
    print STDERR "#    3 : expid_sidG_scanid_YYYYDDDHHMMSS.k5 \n";
    print STDERR "RecordHost sirius-c1 \n";
    print STDERR "RecPortNum 9902\n";
    print STDERR "#RecMode    (Msps):(bit):(nch) \n";
    print STDERR "RecMode    16:1:4 \n";
    print STDERR "NchTotal   16 \n";
    print STDERR "Logdir    /tmp/\n";
    print STDERR "-----------------------------\n";
    exit(1);
}



# use Exporter;
use IO::Socket;                   # IO::Socket ⥸塼Ȥ
use lib "$ENV{HOME}/prog/perl/module";
# use lib "$ENV{HOME}/vdif/MSPerl-module";
use Sched_db;
use Conf_hash;
use Time_lib;
use Term::ANSIScreen qw/:color :cursor :screen :keyboard/;
use Time::HiRes qw(clock_gettime clock_getres CLOCK_REALTIME usleep);


##my @cmds= qw/"in2net=connect:192.42.120.66" "mode=ext:0xff:1" "play_rate=data:08" "net_protocol=UDP:3200000:10" "in2net=on"/;

#my $cnt=0;
my $extra_end = 7;   ## extra delay sec for VSI
my $exflag =0;

{
    usage() if(@ARGV <1);
    ##my $srcport = 9902;
    ##my $dstport = 9901;
    ##my @terms = split(/:/,shift);
    ##my $srchost = $terms[0];
    ##$srcport = $terms[1]  if(@terms > 1);

    ##@terms = split(/:/,shift);
    ##my $dsthost = $terms[0];
    ##$dstport = $terms[1]  if(@terms > 1);
    ##my $msps = shift;
    ##my $bit = shift;
    ##my $nch = shift;
    my ($skdfile, my $conf);
    ##my $exflag =0;
    ##my $fn=2;
    my $sim = 0;
    ##my $Msps=32; # Msps
    ##my $bit = 2; # bit
    ##my $gch = 4; # $channel in a group
    ##my $nch=16;   # Total number of channel
    #my $srchost="localhost";
    ##my $srchost="sirius-c1";
    ##my $srcport=9902;
    my @allstat = ();

    while(@ARGV) {
	my $opt = shift;
	if($opt =~/-skd/){
	    unless(@ARGV){ usage();}
	    $skdfile = shift;
	}elsif($opt =~/-c/){
	    unless(@ARGV){ usage();}
	    $conf = shift;
	}elsif($opt =~/-sim/){
	    $sim = 1;
	}
    }

    unless( defined($skdfile)){	print " E Missing sched file.!!\n"; usage();  }
    unless( defined($conf)){	print " E Missing Config file.!!\n"; usage();  }
        
    my $conf_dat = new Conf_hash;
    die("Can't open a config file: $!")  if ($conf_dat->Read($conf) < 1) ;
    
    my $stid = $conf_dat->Value("StationID");
    print STDERR "E StationID format is wrong: $conf \n" unless(defined($stid));
    print "Stid: $stid \n";

    my $odir  = $conf_dat->Value("Outdir");
    print STDERR "E Outdir format is wrong: $conf \n" unless(defined($odir));
    my @odir;
    my @item = split(/:/,$odir);
    if( @item < 4){ 
	@odir = ($odir,$odir,$odir,$odir);
    }else{  @odir = @item; } 
    
    
    my $fn  = $conf_dat->Value("NamingRule");
    print STDERR "E NamingRule format is wrong: $conf \n" unless(defined($fn));
    print "Naming Rule:$fn\n";

    my $srchost  = $conf_dat->Value("RecordHost");
    print STDERR "E NamingRule format is wrong: $conf \n" unless(defined($srchost));
    print "SrcHost: $srchost\n";
    
    my $srcport  = $conf_dat->Value("RecPortNum");
    print STDERR "E RecPortNum format is wrong: $conf \n" unless(defined($srcport));
    print "SrcHost: $srcport\n";
    
    my $mode  = $conf_dat->Value("RecMode");
    print STDERR "E RecMode format is wrong: $mode \n" unless(defined($mode));
    my @Msps_bit_ch =  split(/:/,$mode);
    if(@Msps_bit_ch < 3){print "E RecMode format is wrong: $mode \n";   }
    print "Modet: $mode\n";
    
    my $nch  = $conf_dat->Value("NchTotal");
    print STDERR "E NchTotal format is wrong: $conf \n" unless(defined($nch));
    print "Ncht: $nch\n";
    
    my $logdir  = $conf_dat->Value("Logdir");
    print STDERR "E Logdir format is wrong: $conf \n" unless(defined($logdir));
    print "Logdir: $logdir\n";
    
    unless($nch % $Msps_bit_ch[2] == 0){
	print "E Total Number of Channel is not multiple of #group channels.\n";
	    exit(1);
    }

    my $skd = new Sched_db;
    $skd->Read($skdfile);
    my $nskd = $skd->Nskd();
    my $srate = $skd->Srate();
    my $expcode = $skd->Expr();

    my @ids = $skd->Stn_ids();
    foreach my $item ( @ids ) {
	chomp($item);
# 	$item=~ s/\r//;
# 	$item=~ s/\n//;
	my $statinc = $skd->Stn_name($item);
	push(@allstat,$statinc);
    } 
    $allst=join(" ",@allstat);
    my $statnm = $skd->Stn_name($stid);
    
    $skdfile =~m!.+/(\S+)\.(\S+)$!;
    my $skdbody = $1;
    my $skdext = $2;
    $logfile = sprintf ("%s/%s_ctl.log", $logdir,$skdbody);
    
    my $ngrp = $nch/$Msps_bit_ch[2];
    
    ## Set Play Rate
    my $rate = $Msps_bit_ch[0]*$Msps_bit_ch[1]*$Msps_bit_ch[2];

    ##$msg = sprintf "play_rate=data:%d;", $rate;
    ##print "Cmd: $msg\n";
    ##&communicate($srchost,$srcport,"$msg\n");
    $exflag = &communicate($srchost,$srcport,"play_rate=data:$rate;");
#    sleep(1);    

    ## Set Mode Ѥʤ
    ##$msg = sprintf "mode=%d:%d:%d;", $Msps, $bit, $nch;
    ##print "Cmd: $msg\n";
    ##&communicate($srchost,$srcport,"$msg\n");
#    sleep(1);    

	## Set Data_out
    &communicate($srchost,$srcport,"data_out=multi:$ngrp;");
    
    ## Set Modegrp
    for(my $i =0;  $i < $ngrp; $i++){
	my $mode = sprintf "%d:%d:%d:%d:%d", $i, $Msps_bit_ch[0],
	$Msps_bit_ch[1], $Msps_bit_ch[2], $i*$Msps_bit_ch[2];
	&communicate($srchost,$srcport,"modegrp=$mode;");
    }

    my $s_size = ` free -m | grep cache | tail -1 | awk '{print \$4}' `;
    
#оݤȤʤStationIDΥ󥹥ȡɼ
    my $c1 = 0; my $c2 = 1;
    my $sv_start_dur=-1; my $sv_last_dur=-1;
    
    while( $sv_start_dur == -1 && $c1 < $nskd){
	$sv_start_dur = $skd->Skd_dur2($c1,$stid);
	$c1++;
    }
    
    while( $sv_last_dur == -1 && $c2 < $nskd){
	$sv_last_dur = $skd->Skd_dur2(($nskd-$c2),$stid);
	$c2++;
    }
    if($c2 == $nskd || $c1 == $nskd){
	print STDERR "Station ID($stid) is not in the schedule file\n";
	exit 1;
    }

    my $sv_start_t2 = $skd->Skd_start(($c1-1),3);
    my $sv_last_t2 = $skd->Skd_stop2(($nskd-$c2+1),$stid,3);
    
    my $s_start_srcname = $skd->Skd_src(0);
    my $s_last_srcname = $skd->Skd_src($nskd-1);
    my $s_start_dur = $skd->Skd_dur2(0,$stid);
    
    my $lastsid=$skd->Skd_stns($nskd-1);
    my $last_stid =chop($lastsid);
    my $s_last_dur = $skd->Skd_dur2($nskd-1,$last_stid);
    
    my $s_start_t1 = $skd->Skd_start(0,3);
    my $s_last_t1 = $skd->Skd_start($nskd-1,3);
    
    my $expr = $skd->Expr();
    
    print "\n\n\n";
    print "*****************************************************************\n";
    print "* send_start2 Ver 1.00 (2013-03-14) by miya *\n";
    print "*****************************************************************\n";
    printf "Experiment Code   : %s\n",$expr; #r1267 Freq_G (a)
    printf "Schedule File     : %s (type : SKED)\n",$skdfile;
    printf "Log File          : %s\n",$logfile;
    printf "Stations Included : %s\n",$allst;
    printf "My Station Name   : %s (My Station ID : %s)\n",$statnm,$stid; #Subnet Mode : ON
    printf "Current Out Dir   : %s %dMB\n",$odir[0],$s_size; #/k56a/ad4/ 271629.2MB left
    printf "Total Scan Number : %s\n",$nskd; #1056 Satellite Mode Keys: HYB NOZ GEO
    printf "1st Scan          : %s (%s sec) %s\n",
    $s_start_t1->Str_fmt3(2),$s_start_dur,$s_start_srcname; #2007/03/17 17:00:00 ( 138sec) 3C446
    printf "Last Scan         : %s (%s sec) %s\n",
    $s_last_t1->Str_fmt3(2),$s_last_dur,$s_last_srcname; #2007/03/18 16:59:10 ( 40sec) 1124-186
    printf "     Range (start): %s (%s sec)\n",
    $sv_start_t2->Str_fmt3(2),$sv_start_dur; #2007/03/17 17:00:00 each scan EARLY start : 10 sec
    printf "     Range (end)  : %s (%s sec)\n",
    $sv_last_t2->Str_fmt3(2),$sv_last_dur; #2007/03/18 16:59:50
    printf "Sampling Mode     : %dMHz %dbit %dch * %dgroup\n",
    $Msps_bit_ch[0],$Msps_bit_ch[1],$Msps_bit_ch[2],$ngrp; #for Next Scan (LPF 8MHz)
    

    my $catch_flag=0;
    my $tofst = 0;
    $tnow = time();

    if($sim ){ 
	$t1 = $skd->Skd_start(0,3)->UXtime();
	$tofst = $tnow - $t1 + 5;
    }

    for(my $iscan=0; $iscan < $nskd; $iscan++){
		my $srcname = $skd->Skd_src($iscan);
		my $dur = $skd->Skd_dur2($iscan,$stid);
		&Logout(":Scan:$iscan $srcname $dur sec");
		next if( $dur < 0);
		$t1 = $skd->Skd_start($iscan,3);
		$t2 = $skd->Skd_stop2($iscan,$stid,3);
		$t1 = $t1 + $tofst;
		$t2 = $t2 + $tofst;
		$tnow = clock_gettime(CLOCK_REALTIME);
		my $twait = $t1->UXtime() - $tnow;
		my $twait2 = $t2->UXtime() - $tnow;

		my $tm = new Time_lib;
		if( $twait > 1){
		    print "\n\n" if ($catch_flag==0);
		    $catch_flag=1;
		    while($twait > 0.5){
			print up(2);
			my $tw = $twait > 1 ? 1.e6 : ($twait-0.2)*1.e6;
			$tm->Set_time_now();
			printf "Next Scan:$iscan $srcname: %s (%s sec) waiting (%d sec) for Start \n",
			$t1->Str_fmt3(2),$dur,$twait;#2007/03/17 17:00:00 (40sec) OJ287 [320.0MB]
			printf "Time Now (UTC)    : %s\n",$tm->Str_fmt3(2);
			##printf "Now(%s) waiting (%d sec) for Start:%s\n",
			##	$tm->Str_fmt3(2), $twait,$t1->Str_fmt3(2);
			usleep($tw);
			$tnow   = clock_gettime(CLOCK_REALTIME);
			$twait  = $t1->UXtime() - $tnow;
			$twait2 = $t2->UXtime() - $tnow;
		    }
		}
		#print "$twait2\n";
		if($twait2 < 0){
		    print "$iscan ";
		    next;
		}
		my @fname;
		for(my $igrp=0; $igrp<$ngrp; $igrp++){
		    $fname[$igrp] = $odir[$igrp] . "/"
			. &Naming($skd,$iscan,$stid,$igrp,$fn);
		    my $outspec = sprintf("data_out=FILE%d:%s;",
			$igrp,$fname[$igrp]);
		    ##print "$outspec \n";
		    &communicate($srchost,$srcport,$outspec);
		    
		}

	#	if( $exflag==1 && $cnt>=6 ){
		if( $exflag==1 ){
		    usleep(200000);
		    &communicate($srchost,$srcport,"in2net=on;");
	#	    print "Output in2net=on \n";
		    &Logout(":Msg Output in2net=on");
		    while($twait2 >0){
				my $tw = $twait2 > 1 ? 1.e6 : ($twait2)*1.e6;
				$tm->Set_time_now();
				print up(2);
				printf "Now Scan: $iscan $srcname: %s (%s sec) waiting (%d sec) for End \n",
				$t1->Str_fmt3(2),$dur,$twait2;
				printf "Time Now (UTC)    : %s\n",$tm->Str_fmt3(2);
				##printf "Now(%s) waiting (%d sec) for End:%s   \n",
				##$tm->Str_fmt3(2), $twait2,$t2->Str_fmt3(2);
				usleep($tw);
				$tnow   = clock_gettime(CLOCK_REALTIME);		
				$twait2 = $t2->UXtime() - $tnow + $extra_end;
		    }
		    &communicate($srchost,$srcport,"in2net=off;");
		    &Logout(":Msg Output in2net=off");
		    print up(9);
		    $s_size = ` free -m | grep cache | tail -1 | awk '{print \$4}' `;
		    printf "Current Out Dir   : %s %dMB\n",$odir[0],$s_size; #/k56a/ad4/ 271629.2MB left
		    print down(9);
		}
    }
    
}

sub Naming($$$;$){
    my ($skd, $i, $id, $igrp,$fn) = @_;
    $fn = 1 unless(defined($fn));
    my $t1 = $skd->Skd_start($i,3);
    my $doy = $t1->Doy();
    my @grp = qw(a b c d);
    my @ydhms = $t1->YDHms();
    my $fname;    
    if($fn == 1){
	$fname = sprintf("%s%03d%04d.dat",$id,$doy,$i);
    }elsif($fn == -1){
	$fname = sprintf("%s%03d%04d.%d.dat",$id,$doy,$igrp,$i);
    }elsif($fn == 2){
	$fname = sprintf("%s%03d%02d%02d%02d%s.dat",
			 $id,$ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4],
			 $grp[$igrp]);
    }elsif($fn == -2){
	$fname = sprintf("%s%03d%02d%02d%02d%s.%d.dat",
			 $id,$ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4],
			 $grp[$igrp], $igrp);
    }elsif($fn == 3){
	@grp = qw(1 2 3 4);
	my $expr = lc($skd->Expr());
	my $src =  lc($skd->Skd_src($i));
	my $hm = sprintf("%02d%02d", $ydhms[2],$ydhms[3]);
	my $id2 = lc($skd->Stn_id2($id));
	$fname = sprintf("%s_%s%s_%s_%04d%03d%02d%02d%02d.k5",
			 $expr, $id2, $grp[$igrp],$hm,$ydhms[0],
			 $ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4]);
    }
    return $fname;
}

sub Logout($){
    my  $msg = shift;
    unless(open(OUT, ">>$logfile")){
	print ("Can't open a logfile: $!");
	return -1;
    }
    my $tm_sock = new Time_lib;
    $tm_sock->Set_time_now();
    my  $now_sock = sprintf "%s", $tm_sock->Str_fmt1(2); 

    print(OUT "$now_sock: $msg\n");
    close (OUT);
    return 0;
}


sub communicate($$$){
	
    my ($host,$port,$line) = @_;
    my $find = index ($line, "=");
    my $cmd =substr($line,0,($find));
#     my $tm_sock = new Time_lib;
#     $tm_sock->Set_time_now();
    my $remote = IO::Socket::INET->new(Proto    => "tcp",
				       PeerAddr => $host,
				       PeerPort => $port );
#    $now_sock = sprintf "%s", $tm_sock->Str_fmt1(2);
    
    ##if( ! $remote){	die "Could not connect to $host:$port \n";    }
    if( ! $remote){	
	&Logout("Error 'Could not connect to $host:$port'");
# 	open(OUT, ">>$logfile") || die("Can't open a logfile: $!"); 
# 	print(OUT "$now_sock:Error 'Could not connect to $host:$port' \n");
# 	close (OUT);
##	$exflag=0;
	return -1;
    }else{$exflag=1;}
	
    select($remote); $|=1; # stop buffering 
    select(STDOUT); $|=1; # stop buffering

    ##print STDERR "Sent: $line\n";
    unless(defined($remote)){
	print STDERR "E Communicate(): Failed to connect host($host)\n";
	return;
    };
    print $remote "$line\n";

    my $res = <$remote>;
    unless(defined($res)){
	print STDERR "E Communicate(): Failed to Read\n";
	return ;
    }
    chomp($res);
    ##print STDERR "Res[$cnt]: $res\n" if(defined($res));
    chomp($res);
#    &Logout(":$line\nk:#$cmd#$res"); 
    &Logout(":$line");
    &Logout(":#$cmd#$res"); 
#     open(OUT, ">>$logfile") || die("Can't open a logfile($logfile): $!");
#     print(OUT "$now_sock:$line\n$now_sock:#$cmd#$res \n"); 
#     close (OUT);
    
#    $cnt++;
    close($remote);
    
    return $exflag;
}
