/** 
 * @Author M.Sekido
 * @Version $Rev: 56 $
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: sudp_recv.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */
#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include "VDIF_Frame.hpp"
#include "SUDPProto.hpp"
#include "Raw_Proto.hpp"
#include "InNet_mdl.hpp"
#include "Dumper.hpp"
#include "conf_utl.hpp"

#include <boost/algorithm/string.hpp>

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

/** General Configulation data */
Config_set vdif_conf;


void usage(){
    cerr << "Usage: " << me <<" [-a <host>][-p <port>] [-m <mode>] [-o <file>]"
	 << endl;
    cerr << "  [-s <byte>] [-l <sec>] \n";
    cerr << "  [-r <Mbps>] [-d <logging file>]\n";
    cerr << "  ----------------------------------------------------\n";
    cerr << " -a <host>, Server host.\n";
    cerr << " -p <port>, default=9901\n";
    cerr << " -m <mode>  Sampling mode. ex) 32Msps/2bit/16ch,  Default: 16Msps/1bit/4ch.\n";
    cerr << " -o <file> Output to file\n";
    cerr << " -s <byte> Packet size (default=1500).\n";
    //  cerr << " -r <Mbps> Expected rate (default=100).\n";
    cerr << " -l <sec> length of receiving data(default=0 means endless).\n";
    cerr << " -d <file> File Logging for Debugging.\n";
    cerr << "-------------------------------------------------------\n";
    cerr << "  If [-o <outfile>] is not specified, data is just receiving to"
	 << "memory and no output." << endl;
    cerr << "  Receivind SUDP stream and can be recorded on file in VDIF format." 
	 << endl; 
    exit(1);
}

// int aflag=0;
// void alarm_signal (int sig) {
//   aflag=1;
//   return;
// }  


int main(int argc, char *argv[]){
    me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 
    int port(DEFAULT_SUDP_PORT);
    string ip("0.0.0.0");
    string fname;
    vector <string> host_list(0);
    int mtu(DEFAULT_MTU_SIZE);  
    uint32_t length(0);
    //  int rate(900);
    string hostname;
    string mode = "16Msps/1bit/4ch";
  
    int k=0;
    while (k >= 0) { 
	switch (k = getopt(argc, argv, "ha:p:o:s:l:d:m:")) { 
	case 'p':   port  = atoi(optarg); break;
	case 'a':   hostname  = optarg; break;
	case 'm':   mode = optarg; break;
	case 's':   mtu  = atoi(optarg); break;
	    //    case 'r':   rate  = atoi(optarg); break;
	case 'l':   length  = atol(optarg); break; // sec
	case 'o':   fname  = optarg; break;
	case 'h':    usage(); break;
	case '?':    usage(); break;
	}
    }
    string logdir = "/tmp/";
    string info_file = logdir + me + "_info.txt";
    string debug_file = logdir + me + "_dbg.txt";
    string error_file = logdir + me +"_error.txt";
    Logsystem::info_logger()->set_stream(info_file);
    Logsystem::debug_logger()->set_stream(debug_file);
    Logsystem::error_logger()->set_stream(error_file);

    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::debug_logger()->enable_logging();

    cerr << " Accepting SUDP packet in VDIF format at port:" << port << endl;
    DAS_Info   das_inf;
  
    try{
	int pos = mode.find("Msps/"); 
	if(pos < 0) throw  "Input mdoe missing 'Msps/'";
	string work = mode.substr(pos+5);
	int Msps = atoi(mode.substr(0,pos).c_str());
	pos = work.find("bit/"); 
	if(pos < 0) throw  "Input mdoe missing 'bit/'";
	int bit = atoi(work.substr(0,pos).c_str());
	work = work.substr(pos+4); 
	pos = work.find("ch");
	if(pos < 0) throw  "Input mdoe missing 'ch'";
	int nch = atoi(work.substr(0,pos).c_str());
	das_inf.Set_Param(nch, bit, Msps*1000000);  
    }
    catch(const char* msg){
	cerr << msg << endl;
	usage();
    }

    VDIF_Frame vdif_frame;
    vdif_frame.Set_DasInfo(das_inf);
    cerr << "Intial DAS = " << das_inf.Mode() << endl;

    SUDPProto sudp(mtu);
    sudp.Set_VlbiFrame(&vdif_frame);
    sudp.Set_DataPort(port);
    sudp.Set_TransLength(length); // sec
    //  sudp.Set_Speed(rate); // (Mbps)
    sudp.Disp_RxRate();
  
    InNet_Model innet;
    innet.Set_protocol(&sudp);
    innet.Set_VlbiFrame(&vdif_frame);
    innet.Set_DasInfo(das_inf);

    if(fname.empty()){ // Just receiving to memory 
	if( innet.Ready() < 0 ){
	    cerr << "Error in innet.Ready()\n";
	    exit(1);
	}
	innet.start();
	innet.join();
    }else{
	Dumper dumper;

	RAW_Proto raw_pro;
	VDIF_Frame vdif;
	innet.Set_VlbiFrame(&vdif);
	vdif.Set_DasInfo(das_inf);
	raw_pro.Set_Host(fname.c_str());
	raw_pro.Set_VlbiFrame(&vdif);
  
	if(dumper.Set_IO(&raw_pro, &innet) <0){
	    cerr << "Error Dumper::Set_IO()\n";
	    return -1;
	}
  
	if( dumper.Ready() ){
	    cerr << " Dump Starting." << endl;
	    dumper.start();
	}else{
	    cerr << " Dump::Ready() Error." << endl;
	}
    
	dumper.join(); // Waiting for the dumper-thread returns.
    }
//   char sRecvBuf[256];
//   while(1){
//     fprintf(stderr,"Terminate by \"QUIT\" \n");
//     scanf("%s", sRecvBuf);
//     StrUpper(sRecvBuf,strlen(sRecvBuf));
//     if(strncmp(sRecvBuf,"QUIT",4)==0){
//       break;
//     }
//   }

    cerr << "Recerving End.\n";
  
}


