/** 
 * @Author M.Sekido
 * @Version $Rev: 56 $
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: sudp_recv2.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */
#include <iostream>
#include <string>
#include <vector>
using namespace std;
#include <boost/algorithm/string.hpp>


#include "VDIF_Frame.hpp"
#include "SUDPProto.hpp"
#include "Raw_Proto.hpp"
#include "InNet_mdl.hpp"
#include "Dumper.hpp"
#include "StaticVals.hpp"
#include "conf_utl.hpp"

char *me;  // My name
/** General Configulation data */
Config_set vdif_conf;

void usage(){
  cerr << "Usage: " << me <<" [-p <port>] [-o <file1>[:<file2>..]] \n";
  cerr << "  [-t <FMT>] [-n <nch>] [-g <grp1>[:<grp2>...]] [-s <byte>] \n";
  cerr << "  [-m <outmode>] [-M <inmode>] [-l <sec>] [-v <n>] [-r]\n";
  //  cerr << "  [-r <Mbps>]  [-d <logging file>]  [-a <host>]\n";
  cerr << "  ----------------------------------------------------\n";
  cerr << "  Receiving data VDIF/SDUP stream, and store the data\n";
  cerr << " -r : Passive recoding mode. Output mode is determined as it is received. \n";
  cerr << "      Othewise 16Msps/1bit/4ch is default for recording mode.\n";
  cerr << "      If -m option is given with -r option, the last option is used.\n";
  cerr << " -p <port>, Data receiving port number. default=9901\n";
  cerr << " -o <file>[:<file2>..] Output to files for multiple out.\n";
  cerr << " -t <FMT>: Writing format, VDIF | VSSP. (default:VSSP32).\n";
  cerr << " -n <nch>, Number of channels (1|4) for output. default=4.\n";
  cerr << " -g <group>[:<grp>...], Writing Group number of nch. 0,1,2..\n";
  cerr << " -m <'xxMsps/xbit/xxch'>: Output Data mode\n";
  cerr << " -M <'xxMsps/xbit/xxch'>: Input  Data mode\n";
//   cerr <<"        Msps -> [04Msps|08Msps|16Msps|32Msps|64Msps]\n";
//   cerr << "        bit  -> [1bit|2bit|4bit|8bit] \n";
//   cerr << "        ch   -> [16ch|08ch|04ch|02ch|01ch] \n";
//   cerr << " -m <mode> Recording mode. ex) 32Msps/2bit/16ch, Default:16Msps/1bit/4ch.\n";
  cerr << " -s <byte> Packet size (default=1500).\n";
  //  cerr << " -r <Mbps> Expected rate (default=100).\n";
  cerr << " -l <sec> length of receiving data(default=0 means endless).\n";
  cerr << "  -v: verbose level, 0--4 (default -1: no logging).\n";
  cerr << "-------------------------------------------------------\n";
  cerr << "  If [-o <outfile>] is not specified, data is just receiving to"
       << "   memory and no output.\n";
  cerr << " Upgarad verion of sudp_recv.\n";
  cerr << " Supporting format: VDIF, VSSP32.\n";
  cerr << "-------------------------------------------------------" << endl;
  exit(1);
}

void LogSetup(const int verbose){

  string logdir = "/tmp/";
  
  string error_file = logdir + me +"_error.txt";

   string debug_file = logdir + me + "_dbg.txt";
   Logsystem::debug_logger()->set_stream(debug_file);
   string info_file = logdir + me + "_info.txt";
   Logsystem::info_logger()->set_stream(info_file);
   string warn_file = logdir + me + "_warn.txt";
   Logsystem::warning_logger()->set_stream(warn_file);
   string err_file = logdir + me + "_err.txt";
   Logsystem::error_logger()->set_stream(err_file);
   
  switch(verbose){
  case 0:
    Logsystem::error_logger()->enable_logging();
    break;
  case 1:
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 2:
    Logsystem::debug_logger()->enable_logging(Logger::MEDIUM); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 3:
    Logsystem::debug_logger()->enable_logging(Logger::DETAIL); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  case 4:
    Logsystem::debug_logger()->enable_logging(Logger::ALL); 
    Logsystem::info_logger()->enable_logging();
    Logsystem::error_logger()->enable_logging();
    Logsystem::warning_logger()->enable_logging();
    break;
  default:
    cerr << " No Logging.\n";
    break;
  }
  Logsystem::debug_logger()->Tstamp() << "==== Logging start =====\n";
  Logsystem::debug_logger()->Flush();
  Logsystem::info_logger()->Tstamp() << "==== Logging start =====\n";
  Logsystem::info_logger()->Flush();
  Logsystem::error_logger()->Tstamp() << "==== Logging start =====\n";
  Logsystem::error_logger()->Flush();
  Logsystem::warning_logger()->Tstamp() << "==== Logging start =====\n";
  Logsystem::warning_logger()->Flush();


}

int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 
  int port(DEFAULT_SUDP_PORT);
  string ip("0.0.0.0");
  string fname;
  //  vector <string> host_list(0);
  int mtu(DEFAULT_MTU_SIZE);  
  uint32_t length(0);
  //  int rate(900);
  //  vector <string> hostname;
  string inmode = "16Msps/1bit/16ch";
  string outmode = "16Msps/1bit/4ch";
  enum Format_t {VDIF, VSSP};
  Format_t format = VSSP;
  vector <int> groups;
  int nch(4);
  int verbose=0;
  bool passive_mode = false;

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hp:o:t:n:g:m:s:l:M:v:r")) { 
    case 'p':   port  = atoi(optarg); break;
    case 't':   if(strncmp(optarg,"VDIF",4) ==0) format=VDIF;  break;
    case 'n':   nch = atoi(optarg);  break;
	//    case 'a':   hostname  = optarg; break;
    case 'm':   outmode = optarg; passive_mode = false; break;
    case 'M':   inmode = optarg; break;
    case 'r':   passive_mode = true; break;
    case 'v':   verbose = atoi(optarg); break;
    case 'g': {    
	vector <string> v;
	boost::algorithm::split(v,optarg,boost::is_any_of(":"));
	groups.resize(v.size());
	for(int i=0; i< (int)v.size(); i++) groups[i] = atoi(v[i].c_str());
    } break;
    case 's':   mtu  = atoi(optarg); break;
      //    case 'r':   rate  = atoi(optarg); break;
    case 'l':   length  = atol(optarg); break; // sec
    case 'o':   fname  = optarg; break;
    case 'h':    usage(); break;
    case '?':    cerr << "Missing following parameters\n"; usage(); break;
    case ':':    cerr << "Invalid option\n"; usage(); break;
    }
  }
  
  LogSetup(verbose);

  cerr << " Accepting SUDP packet in VDIF format at port:" << port ;
  if( length >0) cerr << " Rx_Length=" << length << " sec.";
  cerr << endl;

  DAS_Info   outdas;
  DAS_Info   indas;
  
  if(! outdas.SetMode(outmode) || ! indas.SetMode(inmode)){
    cerr << "Error at mode setting\n";
    usage();
  };
  
    
  VDIF_Frame vdif_frame;
//  vlbi_frame->Set_DasInfo(outdas);
//  vdif_frame.Set_DasInfo(indas);

  cerr << "Initial output DASmode = " << outdas.Mode() << endl;
  SUDPProto inp(mtu);
  inp.Set_VlbiFrame(&vdif_frame);
  if(inp.Mtu() > 1500){
    printf("mtu (%d) > 1500 after \n",inp.Mtu()); exit(1);
  }
  inp.Set_DataPort(port);
  inp.Set_TransLength(length+3); // sec
  //  inp.Set_Speed(rate); // (Mbps)
  inp.Disp_RxRate();
  cerr << "Innet Intializing..." << flush;
  InNet_Model innet;
  cerr << " Set Protcol to Innet..." << flush;
  innet.Set_protocol(&inp);

  cerr << " Frame Set..." << flush;
  innet.Set_VlbiFrame(&vdif_frame);
  cerr << " DasInfoe Set..." << flush;
  innet.Set_DasInfo(indas);
  cerr << endl;
  //   if( innet.Ready() < 0 ){
  //     cerr << "Error in innet.Ready()\n";
  //     exit(1);
  //   }
  
  try{
    if(fname.empty()){ // Just receiving to memory 
      if( innet.Ready() < 0 ){
	cerr << "Error in innet.Ready()\n";
	exit(1);
      }
      inp.Set_TransLength(0); // Endless receiving
      innet.start();
      innet.join();
    }else{ // Output to files.
      vector <string> outfname;
      boost::algorithm::split(outfname,fname,boost::is_any_of(":"));
      int ngrp = outfname.size();
      if(groups.size() == 0 ) groups.push_back(0);
      if((int)groups.size() != ngrp) { 
	cerr << "Error, nfiles = "<< ngrp << ":" << fname 
	     << ", but -g specification is not consistent." << endl;
	throw -1;
      }
      Dumper dumper[ngrp];
      RAW_Proto raw_pro[ngrp];
      VLBI_Frame *outframe[ngrp];      
	  
      switch (format){
      case VDIF:
	for(int i=0; i< ngrp; i++){  
	  outframe[i] = new VDIF_Frame;    
	  outframe[i]->Set_DasInfo(outdas);
	  raw_pro[i].Set_TransLength(length); // sec
	}
	Logsystem::info_logger()->Tstamp()
	  << "Data file("<< fname << ") in VDIF format.\n";
	break;
      case VSSP:
	for(int i=0; i< ngrp; i++){  
	  outframe[i] = new K5VSSP32_Frame;     
	  outframe[i]->Set_DasInfo(outdas);
	  raw_pro[i].Set_TransLength(length); // sec
	}
	Logsystem::info_logger()->Tstamp()
	  << "Data file("<< fname << ") in VSSP32 format.\n";
	break;
      default:
	cerr << "Error im main(), unexpected output format \n";
	break;
      }

      for(int i=0; i< ngrp; i++){
	int firstch = groups[i]*nch;
	outdas.SetFirstCh(firstch);
	outframe[i]->Set_DasInfo(outdas);
	raw_pro[i].Set_VlbiFrame(outframe[i]);
	raw_pro[i].Set_Host(outfname[i].c_str());
	if(passive_mode)
	  dumper[i].SetDasMode(Dumper::IN2OUT);
	else
	  dumper[i].SetDasMode(Dumper::USR_OUT_SET);

	if(dumper[i].Set_IO(&raw_pro[i], &innet) <0){
	  cerr <<  "Error: Dumper[" << i  <<  "]::Set_IO()\n";
	  throw -1;
	}
	if(! dumper[i].Ready() ){
	  cerr << " Dump::Ready() Error." << endl;
	  for(int j=0; j<=i; j++) delete outframe[j];
	  throw -1;
	}
	cerr << " Dump Starting." << endl;
	dumper[i].start();
      } // For loop 
	  
      for(int i=0; i<ngrp; i++){
	dumper[i].join(); // Waiting for the dumper-thread returns.
	delete outframe[i];
	Logsystem::info_logger()->Tstamp() 
	  << "I dumper[" << i << "] joined\n";
      }
    } // else end 
    sleep(1);
  } // try end 

  catch(string msg){
    cerr << " Error:" << msg << endl;
  }
  catch(const int ret){
    cerr << "Exit loop due to error" << endl;
  }
  catch(const char* msg){
    cerr << " Error-msg:" << msg << endl;
  }
  cerr << "Recerving End.\n";

  Logsystem::debug_logger()->Flush();  
  Logsystem::info_logger()->Flush();  
  Logsystem::error_logger()->Flush();  
  Logsystem::warning_logger()->Flush();  
}


