/** 
 * @Author M.Sekido
 * @Version $Rev: 60 $
 * @Date  $Date: 2013-07-06 18:43:45 +0900 (Sat, 06 Jul 2013) $
 * $Id: vsi_localwrite_srv.cpp 60 2013-07-06 09:43:45Z sekido $ 
 */
#include <iostream>
#include <vector>
#include <string>

#include <cstdio>
#include <cstring>
#include <cstdlib>
//#include <boost/algorithm/string.hpp>
using namespace std;

#include "SUDPProto.hpp"
#include "Raw_Proto.hpp"
#include "Mk5B_Frame.hpp"
#include "VDIF_Frame.hpp"
#include "InFile_mdl.hpp"
#include "InMem_mdl.hpp"
#include "Trans_ctl.hpp"
#include "InVSI_mdl.hpp"
#include "StaticVals.hpp"
#include "conf_utl.hpp"

extern "C"{
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
}
char *me;  // My name

/** General Configulation data */
Config_set vdif_conf;
/** Transport Control object */ 
TransCtl transctl_proto;

void usage(){
  cerr <<  "Usage: " << me 
       << " [-f <filename> | -m | -d ]  [-p <server port>]\n";
  cerr << " [-t <VDIF | RAW | K5VSSP | MK5B>] [-i <input mode>] [-o <Nout>]\n";
  cerr << " [-w <sec>] [-v <level>] [-x <Normal|Mk5Geo>] [-s <Station Name>:<StnID>]\n";
  cerr << "  [-c <config>] \n";
  cerr << "-----------------------------\n";
  cerr << "  ==DATA SOURCE ==\n";
  cerr << "  -f: Reading K5VSSP file for instead of VSI-H input.\n";
  cerr << "  -m: Reading from memory as dummy data for <sec> duration.\n";
  cerr << "  -d: Input device (PC/VSI).  \n";
  //  cerr << "  -T: Data Frame Type for Input  (default: K5/VSSP32F) \n" ;
  cerr << "  ==Data Output==\n";
  cerr << "  -t: Output Data Frame Type [VDIF/K5VSSP/MK5B] (default: VDIF) \n" ;
  cerr << "  -o: Number of Output Streams (Default=4) \n" ;
  //  cerr << "  -O: Output Media, [FILE/NET]. (default=NET by sudp) \n" ;
  cerr << "  -w: Page Reading Wait time (default: -1 endless) \n" ;
  //  cerr << "  -r: Rate Mbps (default: 100) \n" ;
  cerr << "  ==Control==\n";
  cerr << "  -p: Control Port of this host (default " << DEFAULT_SUDP_PORT
       << ") \n";
  //  cerr << "  -d: log directory path (default /tmp )\n";

  cerr << "  -i: Input mode in two sorts of formats, ex.)'16Msps/2bit/4ch' or '16:2:4'.\n";
  cerr << "    Default(VSSP): using header info.\n";
  cerr << "  -s: Station Name in File header must be within 8 chars. (default 'Kashim34').\n";
  cerr << "      Station ID must be 2 character\n";
  cerr << "  ==Option==\n";
  cerr << "  -x: Extraction mode Normal(default) or Mk5Geo(stream 16-19/32 are LSB(14,15ch).\n";
  cerr << "  -v: verbose level, 0--4 (default -1: no logging).\n";
  cerr << "  -c: Config file, Default:/usr/local/etc/vdif.conf. \n";
  cerr << "      If it does not exist, current Owner/Group id is applied for recording file and log files.\n";
  cerr << "  ----------- $Rev: 60 $ Updated: $Date: 2013-07-06 18:43:45 +0900 (Sat, 06 Jul 2013) $ by $Author: sekido $ $ -------" << endl;
  exit(1); 
}
 

void sint(int num){
  transctl_proto.Terminate();
}

void LogSetup(const int verbose=-1){

  string logdir = "/tmp/";
  //  string error_file = logdir + me +"_error.txt";
   
  string debug_file = logdir + me + "_dbg.txt";
  Logsystem::debug_logger()->set_stream(debug_file);
  string info_file = logdir + me + "_info.txt";
  Logsystem::info_logger()->set_stream(info_file);
  string warn_file = logdir + me + "_warn.txt";
  Logsystem::warning_logger()->set_stream(warn_file);
  string error_file = logdir + me + "_err.txt";
  Logsystem::error_logger()->set_stream(error_file);

  uid_t vdif_uid(0);
  gid_t vdif_gid(0);
  bool chown_flag = false;
  
  if(vdif_conf.Check_Key("Owner") && vdif_conf.Check_Key("Group")){
    vdif_uid = atoi(vdif_conf.Value("Owner").c_str());
    vdif_gid = atoi(vdif_conf.Value("Group").c_str());
    if(vdif_uid * vdif_gid > 0){
      chown_flag = true;
    }else{
      cerr << "E Vdif_Config was used but 'Owner' or 'Group' Entry is missing, thus not used" << endl; 
      chown_flag = false;
    }
  }
  
  int err_arr[4] = {0,0,0,0};
  vector <string> flist;
  flist.push_back(error_file);
  flist.push_back(info_file);
  flist.push_back(debug_file);
  flist.push_back(warn_file);

  switch(verbose){
  case -1:
    cerr << " No Logging." << endl;
    break;
  case 0:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging( verbose );

    err_arr[0] = chown(error_file.c_str(), vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),  vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(), vdif_uid, vdif_gid);
    
    break;
  case 1:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 2:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 3:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 4:
  default:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  }

  for(int i=0; i<4; i++){
    if(err_arr[i] < 0) fprintf(stderr,"E Failed to change Owner [%s]\n", flist[i].c_str());
  }

}

int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 
  string fname;
  string hostname("localhost");
  int srv_port(DEFAULT_CTRL_PORT);
  bool mflag(false); ///< Flag of memory data
  int mtu(DEFAULT_VDIF_FRAMESIZE +8);  
  int Mbps(100); // Mbps 
  int length(86400); // sec
  int Rate_adj_f = 200; // (Hz)

  VLBI_Frame* in_vlbi_frame(0);
  //  DAS_Info das_inf;
  int input_wait=-1;
  int verbose_level = -1;
  string device_type;
  DAS_Info indas( 32000000, 16, 2 );
  string inMode;
  bool   mk4geo_mode (false);
  string Stname="Kashim34", Stid="kb";
  int Nout_strm=4; // Number of Output Streams. 
  const int outmedia(1); // Output media, 0=NET, 1=File
  VLBI_Frame::VLBI_Frame_t frame_type(VLBI_Frame::VDIF_frm);
  string config_file(DEFAULT_CONF_FILE);

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hc:f:dp:mt:w:v:i:x:s:o:R:")) { 
    case 'h':   usage();     break;
    case 'f':   fname = optarg; break;
    case 'c':   config_file = optarg; break;
    case 's':   {
      char *cp = strchr(optarg, ':');
      if(cp ==0) Stname = optarg;
      else{
	*cp='\0'; cp++; Stname = optarg; Stid = cp;
      }
      break;

    }
    case 'd':   device_type = "VSI"; break;
      //    case 's':   mtu = atoi(optarg); break;
    case 't':   
//       if(strncmp(optarg,"VDIF",4)==0){frame_type = VLBI_Frame::VDIF_frm; } 
//       if(strncmp(optarg,"VSSP",4)==0){frame_type = VLBI_Frame::K5VSSP_frm; } 
//       if(strncmp(optarg,"RAW",3)==0){frame_type = VLBI_Frame::K5VSI_frm; } 
//       if(strncmp(optarg,"MK5B",3)==0){frame_type = VLBI_Frame::MK5B_frm; } 

      if(strstr(optarg,"VDIF")!=NULL){frame_type = VLBI_Frame::VDIF_frm; } 
      if(strstr(optarg,"K5VSS")!=NULL){frame_type = VLBI_Frame::K5VSSP_frm; } 
      if(strstr(optarg,"RAW") !=NULL){frame_type = VLBI_Frame::K5VSI_frm; } 
      if(strstr(optarg,"MK5B")!=NULL){frame_type = VLBI_Frame::MK5B_frm; } 
      break;
    case 'w':   input_wait = atoi(optarg); break;
    case 'o':   Nout_strm = atoi(optarg); break;
      //    case 'O':    if(strncmp(optarg,"FILE",4)==0) outmedia=1; break;
    case 'i':   inMode = optarg; break;
      //    case 'r':   Mbps = atoi(optarg); break;
    case 'R':   Rate_adj_f = atoi(optarg); break;
    case 'm':  // length  = atoi(optarg); 
      mflag = true; break;
      //    case 'a':   hostname = optarg; break;
    case 'p':   srv_port  = atoi(optarg); break;
    case 'x':   if(strncmp(optarg,"Mk5Geo",6) == 0) 
	mk4geo_mode=true; break;
    case 'v':   verbose_level = atoi(optarg); break;
      //    case 'l':   length  = atoi(optarg); break;
    case '?':    usage(); break;
    }
  }

  if(mflag == false && fname.empty() && device_type.empty()){
    cerr << "Error: Input Filename/device/memory is not specified.\n";
    usage();
  }

  if(  vdif_conf.Read(config_file.c_str()) <0){
    fprintf(stderr,"======================================================\n");
    fprintf(stderr, "W  Configfile[%s] could not be read.\n", config_file.c_str());
    fprintf(stderr, "   Current Ownership will be used for recording data and log files.\n");
    fprintf(stderr,"======================================================\n");
  }

  
//   if( out_vlbi_frame == 0 ) out_vlbi_frame = new VDIF_Frame;
//   out_vlbi_frame->Set_StnameID(Stname.c_str(),Stid.c_str());
//   string retname;
//   out_vlbi_frame->Get_Stname(retname);
//   cerr << "Tx Frame Station name=" << retname << endl;

  indas.SetMode(inMode);
//   if( ! inMode.empty() ){
//     int pos = inMode.find("Msps/"); 
//     int sr = atoi(inMode.substr(0,pos).c_str());
//     indas.Set_Srate(sr*1000000);
//     string work = inMode.substr(pos+5);
//     pos = work.find("bit/"); sr = atoi(work.substr(0,pos).c_str());
//     indas.Set_Qbit(sr);
//     work = inMode.substr(pos+4);
//     pos = work.find("ch"); sr = atoi(work.substr(0,pos).c_str());
//     indas.Set_Nch(sr);
//   }
  
  cerr << "***Input Mode=" << indas.Mode() << " ***" << endl;
  cerr << "Verbose level= " << verbose_level << endl; 
  LogSetup(verbose_level);
  
  VLBI_Proto* vlbip[Nout_strm];

  
//   RAW_Proto rawp[Nout_strm];
//   SUDPProto sudp[Nout_strm];
  VLBI_Frame* out_vlbi_frame[Nout_strm];
 
  cerr << "Output to FILE\n";
  for(int i=0; i<Nout_strm; i++){
    switch( frame_type)
      {
      case VLBI_Frame::K5VSSP_frm:
	out_vlbi_frame[i] = new K5VSSP32_Frame; 
	cerr << "["<< i <<"]Output Frame: K5VSSP32 Frame\n";
	break;
      case VLBI_Frame::K5VSI_frm: 
	out_vlbi_frame[i] = new K5VSI_Frame; 
	cerr << "["<< i <<"]Output Frame: K5VSI (Raw)Frame\n";
	break;
      case VLBI_Frame::MK5B_frm:  
	out_vlbi_frame[i] = new Mk5B_Frame;
	cerr << "["<< i <<"]Output Frame: Mk5B Frame\n"; 
	break;
      case VLBI_Frame::VDIF_frm:
      default:
	out_vlbi_frame[i] = new VDIF_Frame; 
	cerr << "["<< i <<"]Output Frame: VDIF Frame\n";
	break;
      }
    out_vlbi_frame[i]->Set_StnameID(Stname.c_str(),Stid.c_str());
//     string retname;
//     out_vlbi_frame[i]->Get_Stname(retname);
//     cerr << "Tx Frame Station name=" << retname << endl;

    if(outmedia ==1)  { // Output to file
      vlbip[i] = new RAW_Proto;
      //      cerr << "Output: FILE\n";
    }else {    // Output to Net
      vlbip[i] = new SUDPProto;
      //      cerr << "Output: Network\n";
    }
    
    vlbip[i]->Set_OprMode(TransInfo::TX); // Transmission mode 
    vlbip[i]->Set_Speed(Mbps,true); // Set 100 Mbps
    
    vlbip[i]->Set_Adj_Freq(Rate_adj_f); // Trx-rate adjusting interval
    vlbip[i]->Disp_TxRate(); 
    vlbip[i]->Set_VlbiFrame( out_vlbi_frame[i] ); // Set Output Frame 
    if(mk4geo_mode){
      vlbip[i]->SetExmode_Mk4Geo(); // Extracting LSB 8,9ch->15,16ch
      cerr << "I: Tx mode- Mk4VSI-Geodesy" << endl;
    }
  }
    

  transctl_proto.Set_CtrlPort(srv_port);
  //  out_vlbi_frame->Set_DasInfo(&das_inf); 
  //  in_vlbi_frame->Set_DasInfo(&das_inf); 
  
  //  transctl_proto.Set_DasInfo(&das_inf); // Set DasInfo

  transctl_proto.Set_ApplProto(vlbip, Nout_strm); // Set TransInfo
  //  transctl_proto.Set_ApplProto(vlbip); // Set TransInfo
  
  fprintf(stderr,"Server- Port: %d  MTU= %d \n",  srv_port, mtu);
//   fprintf(stderr,"Initial Rate: %d Mbps For %d sec, Rate Adjust by %d cycle\n",
// 	  Mbps, length, Rate_adj_f);
  /***************************************/

  InDev_Model *input_mdl(0);
  try{
    /** Data Input Device/File */
    if(! fname.empty()){   // sending from file.
      input_mdl = new InFile_Model;
      in_vlbi_frame = new K5VSSP32_Frame;
      in_vlbi_frame->Set_DasInfo( indas );
      input_mdl->Set_VlbiFrame(in_vlbi_frame);
      ((InFile_Model*)input_mdl)->Set_infile(fname.c_str());  
      fprintf(stderr,"Data source : File(%s)  \n", fname.c_str()); 
    }else if(!device_type.empty()){ // Data input from Device
      //    if( strncmp(device_type.c_str(),"VSI",3) == 0) 
      input_mdl = new K5_VSI[Nout_strm];
      in_vlbi_frame = new K5VSI_Frame[Nout_strm];
      for(int i=0; i<Nout_strm; i++){
	((K5VSI_Frame*)in_vlbi_frame +i)->Set_DasInfo( indas );
	((K5_VSI*)input_mdl+i)->Set_VlbiFrame( ((K5VSI_Frame*)in_vlbi_frame)+i );
      }
      //if( strncmp(device_type.c_str(),"VSSP",4) == 0) input_mdl = new K5_VSSP;
      fprintf(stderr,"Data source : Device(%s)  \n", device_type.c_str()); 
      if( input_mdl->TestDev() < 0){
	throw "Failed to open K5/VSI device file";
      }
    }else{ // data from memory
      input_mdl = new InMem_Model(length);
      in_vlbi_frame = new K5VSI_Frame;
      in_vlbi_frame->Set_DasInfo( indas );
      input_mdl->Set_VlbiFrame(in_vlbi_frame);
      fprintf(stderr,"Data source: Memory   \n") ;
    }
    //  input_mdl->Set_PageWait(input_wait);
    input_mdl->Set_StnameID(Stname.c_str(),Stid.c_str());
    if(!device_type.empty()){ // Data input from K5VSI Device
      for(int i=1; i<Nout_strm; i++){
	((K5_VSI*)input_mdl +i)->Set_StnameID(Stname.c_str(),Stid.c_str());
      }
    }

    transctl_proto.Set_InputDev(input_mdl); // Set Input device 
    if(! transctl_proto.SetupCheck()) exit(1);

    input_mdl->Ready();
    input_mdl->start(); // reading data Started 

    //  signal(SIGINT,SIG_IGN);
    if(signal(SIGINT, sint) == SIG_ERR){
      throw "E Failed signal(SIGINT, sint)";
      exit(1);
    };

    transctl_proto.start(); // Transaction server starting
  
    
    cerr << "TransCtrl Thread ...";
    transctl_proto.join();
    cerr << "joined\n";

    for(int i=0; i<Nout_strm; i++)  vlbip[i]->Disp_RateOff();  
    fprintf(stderr,"Disp RateOff\n");
    /** Memory Release */
//     if( input_mdl != 0 ){
//       if(!device_type.empty()){ // Data input from K5VSI Device
// 	fprintf(stderr,"Delete [] input_mdl..");
// 	delete [] input_mdl;
// 	fprintf(stderr,"Delete [] input_mdl\n");
//       }else{
// 	fprintf(stderr,"Delete input_mdl...");
// 	delete input_mdl;
// 	fprintf(stderr,"Delete input_mdl\n");
//       }
//     }
    
//     if( in_vlbi_frame != 0 ){
//       fprintf(stderr,"Delete in_vlbi_frame..");
//       delete in_vlbi_frame;
//       fprintf(stderr,"Delete in_vlbi_frame\n");
//     }
//     for(int i=0; i<Nout_strm; i++)
//       if( out_vlbi_frame[i] != 0 ){ 
// 	fprintf(stderr,"Delete out_vlbi_frame..");
// 	delete out_vlbi_frame[i]; 
// 	out_vlbi_frame[i]=0;
// 	fprintf(stderr,"Delete out_vlbi_frame\n");
//       }
    /****************************/

  }// End of try
  
  catch(const char* msg){
    cerr << "Error:" << msg << endl;
  }  

  if(Logsystem::Test_Logsys(Logsystem::DEBUG)) 
    Logsystem::debug_logger()->close_stream();
  if(Logsystem::Test_Logsys(Logsystem::INFO))   
    Logsystem::info_logger()->close_stream();
}


