
#ifndef __K5VSI_MDL_HPP__
#define __K5VSI_MDL_HPP__

/** 
 * Input Device (K5_VSI) model Class library Header
 * Succeeding from InDev_Model.
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: InVSI_mdl.hpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include <string>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
using namespace std;

#include "UDPSocket.h"
#include "TCPSocket.h"
#include "FileSocket.h"
#include "Thread.h"
#include "netutl.h"
#include "Logsystem.hpp"
#include "Mutex.hpp"

#include "InDev_mdl.hpp"
#include "bitutl.hpp"
#include "time_lib.hpp"

// #include "m5b.hpp"

extern "C" {
#include <stdint.h>
#include <limits.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>  
#include <sys/socket.h>  
#include <sys/time.h>  
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <pthread.h>
#include <signal.h>
#include "k5vsi.h"
}

// const int MAXSTR =           200; 
//const int DEFAULT_MTUSIZE = 1500;
//const int DEFAULT_NPACKET = 100000;
//const int DEFAULT_SNDR_PORT =    52100;
const int DEFAULT_LENGTH =      10;
// const int SNDBUFSIZE    =  16777216;
// mark5 default ports
//const int DEF_M5CTL_PORT =     2620;
//const int DEF_M5DAT_PORT =     2630;

const int MAX_K5VSI_PAGES  = 10;
const int INTVL_IOCTL_USEC = 2000;  // usec 
const int FRAME_EPOCH_RES_MSEC = 100; // msec

// int make_tab(unsigned char table[256]);
// int make_tab_ni(unsigned char table[256]);
// int make_tab_sgnmag(unsigned char table[256]);

/** 
 * Input model for Data Reding from PC-VSI (K5/VSI) Device. 
 * Succeeding from InDev_Model.
 */

class K5_VSI:public InDev_Model{
public:
  K5_VSI(); ///<  Constructor 
  ~K5_VSI(); ///< Destructor 
  const int OpenDev(const DevCheckLevel_t dclevel=STRICT); ///< K5-VSI Device Open 
  const int CloseDev();///< K5-VSI Device Close 
  const int TestDev(); ///< Test device/file Permission  
  /** This is dummy in k5vsi_mdl class. */
  //  const IndevReturns_t MappingData(vector <uint8_t>& arr,time_t epoch);
  /** Getting page address of the epoch, rpage is returning page number. */
  const uint8_t* MappingData(time_t epoch); 
  /// Release memry page address of the epoch, rpage is returning page number.
  const IndevReturns_t  UnMapping(const time_t epoch=0);
  /**  Return page_size(1sec) in 4-byte unit  */
  uint32_t PageLength(){return PageLength(0);}  
  uint32_t PageLength(const int ipage); ///< Returns Pagesize in word
  /// Return device type in string 
  string StrDevType(){string ret("K5/VSI"); return ret;}
  /// Returns Device type 
  InputModelType_t DevType(){return InDev_Model::K5VSI;}
  int InitPage(){return 0;} ///<  Do nothing but for compatibility
  //  const int Set_srate(uint32_t a); ///< Set Sampring rate (sps)
  const int32_t Srate()const; ///< Return Sampling rate (sps)
  //  const int Set_Nch(const int a); ///< Set Nch
  const int Nch()const; ///< Return Nch; 
  //  int Set_Nbit(const int a); ///< Set Quantization bits
  const int Nbit()const; ///< Return Quantization bits
  //  void Set_Length(const char* a); ///< Set Recording duration
  //  void Set_Length(const int a){length=a;} ///< Set Recording duration
  //  const uint32_t Length()const {return length;} ///< Return Recording Duration
  const int Set_Mode(const char* a); ///< Set Observation mode
  const int ISet_Mode();  ///< Setting Observation mode from (srate,nch)
  string Mode(); ///< Return Obseravtion mode
  //  void Set_iGrp(const int i){igrp=i;} ///< Set group No. from 32/nch groups
  //  const int IGrp(void)const{ return igrp;} ///< Return Selected group Number.
  /// Returns top address of that page
  uint32_t GetPageOffset(const int ipage);
  /// Get Device Epoch from PC-VSI board
  time_t   GetDevEpoch(const bool force_check=false);
  /// Get Epoch of the ipage;
  time_t   GetFrameEpoch(const int ipage, const bool force_check=false)const;
  /// Returning epcoh of the ipg page
  time_t   Epoch(const int ipg)const {
    Logsystem::debug_logger()->Tstamp() << "D K5_VSI::Epoch(): Calling GetFrameEpoch(" << ipg << ")\n";
    return GetFrameEpoch(ipg);}
  /// Get Number of Pages;
  const int Npage();
  /// Set NonStop mode
  const int SetNonStop();
  /// Converting raw data to requested data rate/bit/nch
  int32_t Convert(vector <uint8_t> & arr, int offset);
  int FirstPage(); ///< return page number of earliest epoch
  /// Return Page state in string. 
  string StrPageState(const int lvl=0);
  /// Fill DATA_END to all pages
  void PageEnd(){map_enable=false;} // do nothing
protected:
  static  int          device; ///< file descriptor of VSI device
  static  int             mem; ///< file descriptor of memory device
  //  static time_t device_epoch; ///< Epoch obtained from K5-VSI
  static string vsi_devname; ///< K5-VSI Device file name
  static string mem_devname; ///< K5-VSI Memory Device file name
  static int npage;  ///< Number of available pages;
  //  int igrp;  ///< Group of 32/nch; 
  //  uint8_t enc_table[256]; ///< Encoding Table 
  /** Frame_epch memory to avoid requent access */
  uint8_t *page_ptr;  ///< Page pointer for current reading 
  int rpage;  ///< mapped page number for access  
  bool map_enable; ///<> Mapping Enable state, used for stop running.
  bool dev_opened_flag; ///< Flag indicating the dev is opened for this instance.
  static  int num_opened; //< Number of threads opening of this device;
  static  Mutex k5vsi_mutex; ///< mutex instance
  /**  Frame Epoch Time resolution. Time Update interval used in Tm_mem. */
  static int frame_epoch_res_msec;
  static Tm_mem  frame_epoch[MAX_K5VSI_PAGES]; 
  static Tm_mem  device_epoch; ///< Epoch obtained from K5-VSI
  static uint32_t page_length[MAX_K5VSI_PAGES];  // in [word =4byte]
  static bool     page_offset_setflag[MAX_K5VSI_PAGES];  // 
  static uint32_t page_offset[MAX_K5VSI_PAGES];  // 
};

inline string K5_VSI::StrPageState(const int lvl){

  char buf[1024];
  //   sprintf(buf,"Pointng %d/%d page(%lu). Opened by %d threads. ", 
  // 	  rpage, npage, device_epoch, num_opened);

  string ret;
  if(map_enable) ret = "MAPPING_AVAILABLE ";
  else           ret = "MAP_NOTavailable ";
  
  sprintf(buf,"Mapped pg:%d Opnd by %d thrds.", rpage, num_opened);
  ret += buf;

  if(lvl >0){  
    struct tm tm1;
    char dumbuf[1024];
    for (int i=0; i<npage; i++){
      time_t t_epoch = GetFrameEpoch(i);
      gmtime_r ( &t_epoch, &tm1 );
      strftime ( dumbuf,256,"%Y/%j-%H:%M:%S",&tm1 );
      sprintf(buf," pg[%d](%s)",i, dumbuf);
      //      sprintf(buf," pg[%d](%lu)",i, GetFrameEpoch(i));
      if(Logsystem::Test_Logsys(Logsystem::DEBUG, 2)) 
	Logsystem::debug_logger()->Tstamp() << "D K5_VSI::StrPageState(): Calling GetFrameEpoch(" << i << ")\n";
      ret += buf;
    }
  }

  return ret;
}

inline  time_t   K5_VSI::GetDevEpoch(const bool force_check){
  if( device ==0){
    cerr << "Error in GetDevEpoch(): device is not opened yet.\n";
    Logsystem::error_logger()->Tstamp() 
      << "Error in GetDevEpoch(): device is not opened yet.\n";
    return -1;
  }

  time_t ret(0);

  k5vsi_mutex.lock();
  if(force_check ||
     device_epoch.Check_epoch(frame_epoch_res_msec) >0){;
    
    if(Logsystem::Test_Logsys(Logsystem::DEBUG,Logger::SMALL))
      Logsystem::debug_logger()->Tstamp() << "D K5_VSI: GetDevEpoch().."; 
    //    cerr << "\nD K5_VSI: GetDevEpoch().."; 

    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetDevEpoch(): mutex locked.\n";
    ret = ioctl(device,GET_DEVICE_EPOCH,NULL);
    usleep(INTVL_IOCTL_USEC);
    device_epoch.Set_time(ret);
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetDevEpoch(): mutex UNlocked.\n";
    if(Logsystem::Test_Logsys(Logsystem::DEBUG,Logger::SMALL))
      *Logsystem::debug_logger() << "--GetDevEpoch()..OK\n"; 
    //    cerr << "--GetDevEpoch()..OK" << endl; 
  }
  k5vsi_mutex.unlock();
  
  return device_epoch.Time_t();
}

inline time_t K5_VSI::GetFrameEpoch(const int ipage, const bool force_check)
  const{
  if(ipage>= npage){
    fprintf(stderr,
	    "Error in K5_VSi::FrameEpoch(): requesting page(%d) > npage(%d)\n",
	    ipage,npage);
    Logsystem::error_logger()->Tstamp() 
      << "Error in K5_VSi::FrameEpoch(): requesting page(" << ipage 
      << ") > npage(" << npage << ")\n";
    return 0;
  }

  time_t ret(0);

  k5vsi_mutex.lock();  
  if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
    Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetFrameEpoch(): mutex locked.\n";
  if(force_check ||
     frame_epoch[ipage].Check_epoch(frame_epoch_res_msec) >0){;
    if(Logsystem::Test_Logsys(Logsystem::DEBUG,Logger::SMALL))
      Logsystem::debug_logger()->Tstamp() 
	<< "D K5_VSI: GetFrameEpoch(" << ipage << ")..exec\n" ;
    for(int jpage=0; jpage < npage; jpage++){
      ret =  ioctl(device, GET_FRAME_EPOCH, jpage);
      frame_epoch[jpage].Set_time(ret);
      usleep(INTVL_IOCTL_USEC);
    }
    ret = frame_epoch[ipage].Time_t();
    //    *Logsystem::debug_logger() << "--GetFrameEpoch()..OK\n"; 
  }else{
    ret = frame_epoch[ipage].Time_t();
  }
  if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
    Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetFrameEpoch(): mutex UNlocked.\n";  

  k5vsi_mutex.unlock();
  return ret;
}

inline  uint32_t  K5_VSI::PageLength(const int ipage){
  if(ipage>= Npage()){
    fprintf(stderr,"Error in K5_VSi::PageLength(): requesting page(%d) > npage(%d)\n",
	    ipage,npage);
    return 0;
  }

  if( page_length[ipage] ==0){
    Logsystem::debug_logger()->Tstamp() << "D K5_VSI: PageLength().." ;
    k5vsi_mutex.lock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::PageLength(): mutex locked.\n";
    page_length[ipage] = (uint32_t)ioctl(device,GET_FRAME_CLOCK ,ipage);
    usleep(INTVL_IOCTL_USEC);
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::PageLength(): mutex UNlocked.\n";
    *Logsystem::debug_logger() << "--PageLength()..OK\n"; 
    k5vsi_mutex.unlock();
  }
  
  return page_length[ipage];

}

inline  uint32_t K5_VSI::GetPageOffset(const int ipage){  
  
  if(! page_offset_setflag[ipage]){
    Logsystem::debug_logger()->Tstamp() << "D K5_VSI: Page_Offset()\n"; 
    k5vsi_mutex.lock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetPageOffset(): mutex locked.\n";
    page_offset[ipage] = ioctl(device,GET_FRAME_OFFSET,ipage);
    usleep(INTVL_IOCTL_USEC);
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetPageOffset(): mutex UNlocked.\n";
    page_offset_setflag[ipage] = true;
    k5vsi_mutex.unlock();
  }
  
 return page_offset[ipage];

  
}

inline const int K5_VSI::Npage(){

  if(npage ==0){ // yet npage is not obtailed.
    k5vsi_mutex.lock();
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetNpage(): mutex locked.\n";
    npage = ioctl(device,GET_FRAME_NUMBER,NULL);
    usleep(INTVL_IOCTL_USEC);
    if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
      Logsystem::info_logger()->Tstamp() << "I K5_VSI::GetNpage(): mutex UNlocked.\n";
    k5vsi_mutex.unlock();
  }

  return npage;
}

inline const int K5_VSI::SetNonStop(){

  k5vsi_mutex.lock();
  if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
    Logsystem::info_logger()->Tstamp() << "I K5_VSI::SetNonStop(): mutex locked.\n";
  int ioret = ioctl(device,SET_FRAME_FREEZE,0x7FFFFFFF);
  usleep(INTVL_IOCTL_USEC);
  if(Logsystem::Test_Logsys(Logsystem::INFO,Logger::DETAIL))
    Logsystem::info_logger()->Tstamp() << "I K5_VSI::SetNonStop(): mutex UNlocked.\n";
  k5vsi_mutex.unlock();

  return ioret;
}

#endif

