/** 
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: SUDPProto.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include "SUDPProto.hpp"
#include <cstdio>
extern "C" {
#include <sys/time.h>
}

//**********TX1sec() vector-array version ******************/
int32_t SUDPProto::TX1sec(const time_t t1, vector <uint8_t>&arr){
  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1sec(): started.\n";
  if(  addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1sec(): host/port is not set\n";
    return -1;
  }

  int32_t remain = arr.size();
  //  int seq_size = sizeof(uint64_t);
  int seq_size = sizeof(SudpHdr);
  if(seq_size != SUDP_HEAD_SIZE){
    cerr << "E SUDPProto::Tx1sec(): SUDP header size is not "<< SUDP_HEAD_SIZE
	 << ", but "  << seq_size << endl;
    return -1;
  }
  
  //  int64_t* const pseq = (int64_t*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seq_size;
  vector <uint8_t>::iterator dat_itr = arr.begin();
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize + seq_size;
  sudphdr.SetFramesize(pktsize);

  if( remain % pld_size != 0){
    cerr << "E SUDPProto::TX1sec(): bytesec("<< remain 
	 << ") mod pld(" << pld_size << ") !=0" << endl;
    return -1;
  }
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize  << "\n";
    return -1;
  }
  int npkt = remain / pld_size;
  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  vlbi_frame->Reset_header(); // Initializing seq number of Frame  
  vlbi_frame->Set_Head_par(t1);
  Logsystem::debug_logger()->Tstamp() 
    << "TX1sec() sending " << vlbi_frame->Get_Epoch().Print_time(10) 
    << "with framesize:" << vlbi_frame->Framesize() << "\n";
  Logsystem::debug_logger()->Flush();

  string stname,stid;
  vlbi_frame->Get_Stname(stname);
  vlbi_frame->Get_StnID(stid);
  cerr << " TXing data[" <<   vlbi_frame->Get_Epoch().Print_time(10) 
       << "] Stname=" << stname << " StnID=" << stid << endl;

  //  *pseq = seq;
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){

  sudphdr.SetValid();
  for(int ipkt=0; ipkt < npkt && dat_itr < arr.end(); ipkt++){
    //    memcpy(ppld,pbuf,pld_size);
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    // Set SUDP header for the packet
    memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
    vlbi_frame->Make_frame(frame_ptr, &dat_itr[0], pld_size);
    //    copy(dat_ptr, dat_ptr + pld_size, pld_ptr);
    dat_itr += pld_size;   
    sudphdr++; //    (*pseq)++;

    sudphdr++;
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1sec(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){ perror("E SUDPProto::TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
#ifdef _SLOW_
      if( speed <= 8000000){
#ifdef _MY_SLEEP_
      my_usleep(16.0);
#else
      struct timespec tspec = {0,16000};
      nanosleep(&tspec,0);
#endif
      }
#endif
      if(nwrote < pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1sec(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){perror("E SUDPProto::TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent+=nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    //    if(total_sent > byte_adj){
    if(ipkt != 0 && (ipkt % rate_adj_intvl == 0) ) Rate_Adj();
  }
  
  if( remain !=0){ 
    //     fprintf(stderr, 
    // 	    "Error in SUDPProto::TX1sec():  remian=%ld bytes of %lu bytes \n",
    // 	    (long int)remain, (long unsigned int)nbyte);
    if(remain > 0) return remain;
    else           return -1;
  }
  //  status = IDLE;
  Set_St1sec(); // Set 1sec start time
  //  sent_Mbytes += total_sent*1.e-6;
  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1sec(): " << total_sent
				      << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();
  //  seq = *pseq;
  return total_sent;
}


/** SUDPProto::TX1secSub() -- Transmission of sub-group of channels vector array version
 * Transmitting data of sub-group of the whole data.
 */
int32_t 
SUDPProto::TX1secSub(const time_t t1, vector <uint8_t>&arr,
		     const DAS_Info* indas){

  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1secSub(): started.\n";
  if(  addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1secSub(): host/port is not set\n";
    return -1;
  }

  //  int32_t remain = ExtSubCh(arr,das); // Extracting data for subchannel
  int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
  int32_t remain = arr.size()/ratio;

  //  int seq_size = sizeof(uint64_t);
  int seq_size = sizeof(SudpHdr);
  //  int64_t* const pseq = (int64_t*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seq_size;
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  //  uint8_t* dat_ptr = &preSendBuf[0];
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize + seq_size;
  sudphdr.SetFramesize(pktsize);
  if(remain % pld_size != 0){
    fprintf(stderr,"E SUDPProto::TX1secSub(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    return -1;
  }
  if( (int)packet.size() < pktsize){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1secSub(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize   << "\n";
    return -1;
  }

  vector <uint8_t> databuf(pld_size);
  const uint8_t* dat_ptr = &databuf[0];
  int npkt = remain / pld_size;
  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  vlbi_frame->Reset_header(); // Initializing seq number of Frame  
  vlbi_frame->Set_Head_par(t1);
  Logsystem::debug_logger()->Tstamp() 
    << "TX1secSub() sending " << vlbi_frame->Get_Epoch().Print_time(10) 
    << "with framesize:" << vlbi_frame->Framesize() << "\n";
  Logsystem::debug_logger()->Flush();

  string stname,stid;
  vlbi_frame->Get_Stname(stname);
  vlbi_frame->Get_StnID(stid);
  cerr << " TXing data[" <<   vlbi_frame->Get_Epoch().Print_time(10) 
       << "] Stname=" << stname << "Stid=" << stid << endl;
  
  //  *pseq = seq;
  //  ExtSubCh extch(&arr[0], arr.size(), indas, vlbi_frame->DAS_Inf());
  extch.Setup(&arr[0], arr.size(), indas, vlbi_frame->DAS_Inf());
  sudphdr.SetValid();
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1;
  int32_t nbyte(0);
  for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
    if(Is1bitext)    nbyte = extch.Extract1bit(databuf, pld_size);
    else             nbyte = extch.Extract(databuf, pld_size);

    if( nbyte < 0 ){
      cerr << "E SUDPProto::Tx1secSub(): Failed extch.Extract()"
	   << " err_cnt=" << err_cnt << endl;
      err_cnt--; 
      if(nbyte < -1) err_cnt = -1;
      continue;
    }
    memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);

    //    copy(dat_ptr, dat_ptr + pld_size, pld_ptr);
    //    dat_ptr += pld_size;
    //    (*pseq)++;
    sudphdr++;
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1secSub(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){
	perror("E SUDPProto::TX1secSub():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
#ifdef _SLOW_
      if( speed <= 8000000){
#ifdef _MY_SLEEP_
      my_usleep(16.0);
#else
      struct timespec tspec = {0,16000};
      nanosleep(&tspec,0);
#endif
      }
#endif
      if(nwrote < pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1secSub(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){perror("E SUDPProto::TX1secSub():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent+=nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    //    if(total_sent > byte_adj){
    if(ipkt != 0 && (ipkt % rate_adj_intvl == 0) ) Rate_Adj();
  }
  
  if( remain !=0){ 
    //     fprintf(stderr, 
    // 	    "Error in SUDPProto::TX1secSub():  remian=%ld bytes of %lu bytes \n",
    // 	    (long int)remain, (long unsigned int)nbyte);
    if(remain > 0) return -remain;
    else           return -1;
  }
  //  status = IDLE;
  Set_St1sec(); // Set 1sec start time
  //  sent_Mbytes += total_sent*1.e-6;
  Logsystem::debug_logger()->Tstamp()
    << "SUDPProto::TX1secSub(): " << total_sent
    << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();
  //  seq = *pseq;
  return total_sent;
}


//*********  TX1sec() Pointer version *******************/
int32_t SUDPProto::TX1sec(const time_t t1, const uint8_t *arr, 
			  const int32_t ndat){
  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1sec(): started.\n";
  if(  addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1sec(): host/port is not set\n";
    return -1;
  }

  //    status = SNDNG;
  int32_t remain = ndat;
  //  int seq_size = sizeof(uint64_t);
  int seq_size = sizeof(SudpHdr);
  //  int64_t* const pseq = (int64_t*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seq_size;
  //  vector <uint8_t>::itertator frame_itr = packet.begin() + seq_size;
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  const uint8_t* dat_ptr = &arr[0];
  const uint8_t* dat_end = dat_ptr + ndat;
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize + seq_size;
  sudphdr.SetFramesize(pktsize);
  if(remain % pld_size != 0){
    cerr << "E SUDPProto::TX1sec(): bytesec("<< remain 
	 <<") mod pld(" << pld_size << ") !=0" << endl;
    return -1;
  }

  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize << "\n";
    return -1;
  }
  int npkt = remain / pld_size;
  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  vlbi_frame->Reset_header(); // Initializing seq number of Frame  
  vlbi_frame->Set_Head_par(t1);
  Logsystem::debug_logger()->Tstamp() 
    << "TX1sec() sending " << vlbi_frame->Get_Epoch().Print_time(10) 
    << "with framesize:" << vlbi_frame->Framesize() << "\n";
  Logsystem::debug_logger()->Flush();

  string stname, stid;
  vlbi_frame->Get_Stname(stname);
  vlbi_frame->Get_StnID(stid);
  cerr << " TXing data[" <<   vlbi_frame->Get_Epoch().Print_time(10)  << "] Stname="
       << stname << "Stid=" << stid << endl;
  
  //  *pseq = seq;
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  sudphdr.SetValid();
  for(int ipkt=0; ipkt < npkt; ipkt++){
    //    memcpy(ppld,pbuf,pld_size);
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
    if( dat_ptr +pld_size > dat_end ){
      cerr << "data_ptr exeeding end of buffer, ipkt =" << ipkt << " < " 
	   << npkt << endl;
      break;
    }
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    //    copy(dat_ptr, dat_ptr + pld_size, pld_ptr);
    dat_ptr += pld_size; //   (*pseq)++;
    sudphdr++;

    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1sec(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){ perror("E SUDPProto::TX1sec():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
#ifdef _SLOW_
      if( speed <= 8000000){
#ifdef _MY_SLEEP_
      my_usleep(16.0);
#else
      struct timespec tspec = {0,16000};
      nanosleep(&tspec,0);
#endif
      }
#endif
      if(nwrote < pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1sec(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){perror("E SUDPProto::TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent+=nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    //    if(total_sent > byte_adj){
    if(ipkt != 0 && (ipkt % rate_adj_intvl == 0) ) Rate_Adj();
  }

  if( remain !=0){ 
    fprintf(stderr, "E SUDPProto::TX1sec():  remian=%ld bytes of %lu bytes \n",
	    (long int)remain, (long unsigned int)ndat);
    if(remain > 0) return -remain;
    else           return -1;
  }
  //  status = IDLE;
  Set_St1sec(); // Set 1sec start time
  //  sent_Mbytes += total_sent*1.e-6;
  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1sec(): " << total_sent
				      << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();
  //  seq = *pseq;
  return total_sent;
}


/** SUDPProto::TX1secSub() -- Transmission of sub-group of channels, pointer version
 * Transmitting data of sub-group of the whole data.
 */
int32_t 
SUDPProto::TX1secSub(const time_t t1, const uint8_t *rawarr,
		     const int32_t pagesize, const DAS_Info* indas){
  
  Logsystem::debug_logger()->Tstamp() << "SUDPProto::TX1secSub(): started.\n";
  if(  addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1secSub(): host/port is not set.\n";
    return -1;
  }

  //  int32_t remain = ExtSubCh(rawarr, pagesize, das); // Extracting data for subchannel
  int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
  int32_t remain = pagesize/ratio;

  //  int seq_size = sizeof(uint64_t);
  int seq_size = sizeof(SudpHdr);
  //  int64_t* const pseq = (int64_t*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seq_size;
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize + seq_size;
  sudphdr.SetFramesize(pktsize);
  if(remain % pld_size != 0){
    cerr <<"E SUDPProto::TX1secSub(): bytesec(" << remain << 
      ") mod pld(" << pld_size << ") !=0" << endl;
    return -1;
  }

  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX1secSub(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize  << "\n";
    return -1;
  }
  vector <uint8_t> databuf(pld_size);
  uint8_t* dat_ptr = &databuf[0];
  int npkt = remain / pld_size;
  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);
  vlbi_frame->Reset_header(); // Initializing seq number of Frame  
  vlbi_frame->Set_Head_par(t1);

  Logsystem::debug_logger()->Tstamp() 
    << "TX1secSub() sending " << vlbi_frame->Get_Epoch().Print_time(10) 
    << "with framesize:" << vlbi_frame->Framesize() << "\n";
  Logsystem::debug_logger()->Flush();

  string stname, stid;
  vlbi_frame->Get_Stname(stname);  
  vlbi_frame->Get_StnID(stid);
  cerr << " TXing data[" <<   vlbi_frame->Get_Epoch().Print_time(10) 
       << "] Stname=" << stname << "Stid=" << stid << endl;
  //  ExtSubCh extch(rawarr, pagesize, indas, vlbi_frame->DAS_Inf());
  extch.Setup(rawarr, pagesize, indas, vlbi_frame->DAS_Inf());
  
  //  *pseq = seq;
  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  sudphdr.SetValid();
  int32_t nbyte(0);
  bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1 ;
  for(int ipkt=0; ipkt < npkt && err_cnt >0; ipkt++){
    if(Is1bitext)    nbyte = extch.Extract1bit(databuf, pld_size);
    else             nbyte = extch.Extract(databuf, pld_size);
    if( nbyte < 0 ){
      cerr << "E SUDPProto::Tx1secSub(): Failed extch.Extract()"
	   << " err_cnt=" << err_cnt << endl;
      err_cnt--;
      if(nbyte < -1) err_cnt = -1;
      continue;
    }
    memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, nbyte);
    //    copy(dat_ptr, dat_ptr + pld_size, pld_ptr);
    //    dat_ptr += pld_size;
    //    (*pseq)++;
    sudphdr++;
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1secSub(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){perror("E SUDPProto::TX1secSub():"); err_cnt--; continue;}
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
#ifdef _SLOW_
      if( speed < 8000000){
#ifdef _MY_SLEEP_
      my_usleep(16.0);
#else
      struct timespec tspec = {0,16000};
      nanosleep(&tspec,0);
#endif
      }
#endif
      if(nwrote < pktsize){ 
	fprintf(stderr, "E SUDPProto::TX1secSub(): wrote %d /%d byte\n", 
		nwrote, pktsize);
      }
      if(nwrote <0){perror("E SUDPProto::TX1sec():"); err_cnt--; continue;}
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    //    if(total_sent > byte_adj){
    if(ipkt != 0 && (ipkt % rate_adj_intvl == 0) ) Rate_Adj();
  }
  
  if( remain !=0){ 
    fprintf(stderr, "E SUDPProto::TX1secSub():  remian=%ld bytes of %lu bytes \n",
	    (long int)remain, (long unsigned int)nbyte);
    if(remain > 0) return remain;
    else           return -1;
  }
  //  status = IDLE;
  Set_St1sec(); // Set 1sec start time
  //  sent_Mbytes += total_sent*1.e-6;
  Logsystem::debug_logger()->Tstamp()
    << "SUDPProto::TX1secSub(): " << total_sent
    << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();
  //  seq = *pseq;
  return total_sent;
}


/// Reading data from file and send it to the net 
int32_t SUDPProto::TX(FILE *fp){

  if(  addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E SUDPProto::TX(): host/port is not set\n";
    return -1;
  }
  //  status = SNDNG;
  const int mtu = _sock->get_mtu();
  //  uint8_t packet[mtu];
  packet.resize(mtu);
  //  int64_t *pseq = (int64_t*)&packet[0];
  //  int seqsize = sizeof(uint64_t);
  int seqsize = sizeof(SudpHdr);
  uint8_t *frame_ptr(&packet[0] + seqsize);
  //  int pld_size =  mtu - seqsize;
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize + seqsize;  
  sudphdr.SetFramesize(pktsize);
  const int npld = speed*4/pld_size; ///< Number of Packets to be used
  const int32_t  ndat = npld * pld_size;
  vector <uint8_t> buf(ndat);
  uint8_t *pbuf = &buf[0];
  int nwrote(0);
  int32_t  total_sent=0;
  //  int32_t  intvl_sent=0;
  const int err_lmt = ERROR_CNT_LIMIT;
  int err_cnt=err_lmt;  
  //  struct timeval   et;
  //  char time_str[80];
  //  float avg_rate(0);
  //  float inst_rate(0);
  //  double  dt_plan(0), wait(0);
  //  double speed2 = speed*(double)mtu/pld_size;
  int32_t byte_adj_step = speed/rate_adj_frq*4;
  int32_t byte_adj = byte_adj_step;
  Timer_Start();

  //  gettimeofday(&st,NULL);   
  //  *pseq = seq;
  //  int dat_size(0);
  //  int dmtu;
  int32_t remain(0);
  FirstPacket();
  int nbyte(0);
  
  while(! feof(fp) ){
    int32_t nread = fread(pbuf,1,ndat-remain,fp);
    remain += nread;
    pbuf = &buf[0];
    do{
//       if(remain < pld_size){
// 	int dat_size = Min <int32_t>( (int32_t)pld_size, nread);
// 	memcpy(ppld,pbuf,dat_size);
// 	pbuf += dat_size;
// 	dmtu = dat_size + seqsize;
//       }else{
// 	memcpy(ppld,pbuf,pld_size); 
// 	pbuf += pld_size;
// 	dmtu = pld_size + seqsize;
//       }
      nbyte =  Min <int32_t>( (int32_t)pld_size, remain);
      //      (*pseq) ++;
      memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
      vlbi_frame->Make_frame(frame_ptr, pbuf, nbyte);
      sudphdr++;
      pbuf += nbyte;
      nwrote = _sock->sendto((struct sockaddr *)&addr, sizeof(addr), 
			     (char*)&packet[0], pktsize);
      if(nwrote <0){ 
	perror("E SUDPProto::TX(): Failed to send"); 
	err_cnt--; continue;}
      if(nwrote != pktsize){
	fprintf(stderr, "E SUDPProto::TX(): wrote %d /%d byte \n",
		nwrote, pktsize);
	remain -= (nwrote -seqsize);
      }else{
	remain -= nbyte;
      }
      err_cnt = err_lmt;      

      rate_meas.Add_Snd_Pld(nbyte);
      total_sent += nwrote; 
      if(total_sent > byte_adj){
	Rate_Adj();
	byte_adj += byte_adj_step;
      }
    }while(remain >pld_size);
    memcpy(&buf[0],pbuf,remain);
    pbuf=&buf[0] + remain;
  }
  
  fprintf(stderr, " SUDPProto::TX(*fp) sent: =%ld bytes\n", (long int)total_sent);
  //    status = IDLE;
  LastPacket();
  Timer_Stop();
  cerr << "I SUDPProto::Rxvdif():Timer stopped." << endl;
  Logsystem::info_logger()->Tstamp() 
    << "I SUDPProto::Rxvdif():Timer stopped.\n";

  //  seq = *pseq;
  return total_sent;

}


// #if 0
// int32_t SUDPProto::RX(FILE *fp){

//   status = RCVNG;
//   int mtu = _sock->get_mtu();
//   //  uint8_t packet[mtu];
//   packet.resize(mtu);
//   const int64_t * pseq = (int64_t*)&packet[0];
//   int seq_size = sizeof(int64_t);
//   uint8_t *pbuf = &packet[seq_size];
//   memset(&packet[0],0,sizeof(packet));
//   //  int32_t initial_seq=-1;

//   int pld_size = mtu - seq_size;
//   int nread;
//   const int err_lmt = ERROR_CNT_LIMIT;
//   int err_cnt = err_lmt;
//   int32_t rcv_cnt(0);
//   //, intvl_cnt(0);
//   //  int32_t rcv_cnt_max = rx_length==0 ? (uint32_t)(-1)/mtu : rx_length/mtu;
//   int64_t seq_prev(0);
//   int addr_len = sizeof(addr);
//   struct timeval st, ct;
//   gettimeofday(&st,NULL);
//   int init_flag=0;
//   uint64_t seq0(0);
//   ct=st;

//   while( *pseq >= 0 && err_cnt > 0){
    
//     nread = _sock->recvfrom((struct sockaddr&)addr, addr_len, 
// 			    (char*)&packet[0], mtu);
//     if( nread == 0) continue;
//     if(nread != mtu){
//       fprintf(stderr, "Error in recvfrom(): recevd(%d byte)/expected(%d byte)\n",
// 	      nread,mtu);
//       err_cnt--;
//       continue;
//     }
    
//     if(*pseq <0 || *pseq > 1000000000){
//       fprintf(stderr,"Error sequenc Num=%lld is out of range(1GB)\n", 
// 	      (long long)*pseq);
//       Print_RX_rate(rcv_cnt, seq_prev); 
//       break;
//     }
//     if(init_flag==0){ init_flag=1; seq0=*pseq;}
//     fseek(fp, pld_size*((*pseq - seq0)-1), SEEK_SET);
//     if(fwrite(pbuf,pld_size,1,fp) < 1){ 
//       Logsystem::error_logger()->Tstamp() << "Error in writing in RX()\n";} 
    
//     rcv_cnt++; 
//     //    intvl_cnt++;
//     if(*pseq==0 || seq_prev - (*pseq) > 100){ 
//       //      Print_RX_rate(intvl_cnt, seq_prev);
//       //      intvl_cnt=0; 
//       ndup=n_mis=0;
//       err_cnt=err_lmt; seq_prev = (*pseq);
//     }else{
//       if(*pseq <= seq_prev){
// 	ndup++;
// 	n_mis--;
//       }else{
// 	n_mis += *pseq - seq_prev-1;
// 	seq_prev = *pseq % 10000;
//       }
//     }
//     if(rcv_cnt % 10000 ==0) putc('.',stderr);
//     gettimeofday(&ct,NULL) ;     
//     if( txrx_length >0 && ct.tv_sec > (time_t)(st.tv_sec + txrx_length) )
//       break;
//   }

//   int32_t total_wrote = rcv_cnt*pld_size;
//   return total_wrote;
// }
// #endif


int32_t SUDPProto::RX_UDP(void* buf, const int32_t buf_size){
  
  int mtu = _sock->get_mtu();
  const int n=10;
  //    status = RCVNG;
  //  uint8_t packet[mtu*n];
  packet.resize(mtu*n);
  //  const int64_t * pseq = (int64_t*)&packet[0];
  SudpHdr* psudphdr = (SudpHdr*)&packet[0];
  int seqsize = sizeof(SudpHdr);
  uint8_t *pdat = &packet[seqsize];
  memset(&packet[0],0,sizeof(packet));
  memset(buf,0,buf_size);
  //  int32_t initial_seq=-1;
  int pld_size = mtu - seqsize;
  int nread;
  const int err_lmt = ERROR_CNT_LIMIT;
  int err_cnt = err_lmt;
  int32_t rcv_cnt(0);
  //  , intvl_cnt(0);
  //  int32_t seq_prev(0);
  int addr_len = sizeof(addr);
  uint8_t *b0 = (uint8_t*)buf;
  //  int32_t nseq = *pseq;
  //  int32_t npktsps = (int32_t)((double)speed*4/mtu);
  int32_t pktcnt=0;
  int32_t seq, seq0(0);
  time_t t1, t2;
  time(&t1);
  t2 = t1 + txrx_length;
  bool time_lim = true;
  Timer_Start();
  psudphdr->SetInit();
  
  while( psudphdr->Cmd() != SudpHdr::END 
	 && err_cnt > 0  && time_lim ){
    nread = _sock->recvfrom((struct sockaddr&)addr, addr_len, 
			    (char*)&packet[0], mtu*n);
    if(nread == 0) continue;
    if( nread < 0){
      cerr << "E SUDPProto::RX_SUDP():recvfrom() return:" 
	   << strerror(errno) << endl;
      err_cnt--; continue;
    }
    if(nread != mtu){
      fprintf(stderr, "E SUDPProto::RX_SUDP(): recevd(%d byte)/expected(%d byte)\n",
	      nread,mtu);
      err_cnt--;
      continue;
    }
    //    fprintf(stderr,"DBG: seqnum=%d\n", nseq);
    // if( *pseq ==-1){  Print_RX_rate(rcv_cnt, seq_prev); break;} // status=STOP;
    //    if(*pseq <=0) continue;
    //    ipos = ((nseq) -1)  * pld_size;
    seq = psudphdr->Seq();
    if( pktcnt == 0 || psudphdr->Cmd() == SudpHdr::INIT) seq0 = seq;
    int32_t ipos = (seq - seq0) * pld_size;

    //    ipos = (ipos+pld_size) > buf_size ? 0 :ipos ;    
    if( (ipos+pld_size > buf_size) ){
      cerr << "E SUDPProto::RX(): ipos("<< ipos << ")>bufsize(" 
	   << buf_size << ") or #seq(" << seq  << ") is out of range\n";
      continue;
      // break;
    }
    memcpy(b0+ipos, pdat, pld_size);
    
    err_cnt=err_lmt;
    rcv_cnt++; // intvl_cnt++;

    //    if(*pseq ==0 ||  seq_prev - *pseq > 100 || intvl_cnt > npktsps){ 
    //     if(*pseq ==0 ){ 
    //       //      Print_RX_rate(intvl_cnt, seq_prev);
    //       //      fprintf(stderr,"intvl/npktsps = %d/%d \n",intvl_cnt,npktsps);

    //       //      npktsps = (int32_t)((double)speed*1.e6/8/mtu);
    //       //      fprintf(stderr,"new npktsps=%d\n",npktsps);
    //       //      intvl_cnt=0; 
    //       ndup=n_mis=0;
    //       err_cnt=err_lmt; seq_prev = *pseq;
    //     }else{
    //       if(*pseq <= seq_prev){
    // 	ndup++;
    // 	if(n_mis >0) n_mis--;
    //       }else{
    // 	n_mis += *pseq - seq_prev-1;
    // 	seq_prev = *pseq;
    //       }
    //     }
    rate_meas.Add_Rcv_Pld( pld_size , seq );
    pktcnt++;
    if(txrx_length >0 ){
      time(&t1); time_lim = (t1 > t2);
    }
  }// while loop
  int32_t total_wrote = rcv_cnt*pld_size;
  Logsystem::error_logger()->Tstamp() 
    << "SUDPProto::RX(): wrote (" << total_wrote << " Byte).\n";
  Timer_Stop();
  return total_wrote;
}


/** 
 * Simply Receiving data and write to file with VLBI_frame 
 * 
 */

const uint64_t SUDPProto::RX_SUDP(FILE* fp){
  
  int mtu = _sock->get_mtu();
  int seqsize = sizeof(SudpHdr);  
  //    status = RCVNG;
  //  vector <uint8_t> packet(mtu);
  packet.resize(mtu);
  //  const int64_t* pseq = (int64_t*)&packet[0];
  SudpHdr* psudphdr = (SudpHdr*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seqsize;
  memset(&packet[0],0,packet.size());

  //  int32_t rcv_cnt(0);
  //  int64_t seq_prev(0);
  int addr_len = sizeof(addr);
  //  runflag = true;
  applProtocolState = RXING;
  int pktsize = mtu;
  int framesize = pktsize - seqsize;
  Timer_Start();
  //  int nloop = N_LOOP_LIMIT;
  int32_t pktcnt(0);
  int err_cnt = ERROR_CNT_LIMIT;
  int32_t w_pos =0;
  int32_t seq, seq0(0);
  uint64_t total_wrote(0);

  while( err_cnt > 0 && applProtocolState == RXING 
	 && psudphdr->Cmd() != SudpHdr::END 
	 && (EndEpoch ==0 || time(0) < EndEpoch)){
    int nread = _sock->recvfrom((struct sockaddr&)addr, addr_len,
				(char*)&packet[0], mtu);
    if( nread == 0 ) continue;
    switch(nread){
    case -1:
      perror("E SUDPProto::RX_SUDP():recvfrom() error");
      err_cnt--;
      continue; break;
    case 0:
      cerr << "E SUDPProto::RX_SUDP():recvfrom() no data" << endl;
      continue; break;
    default:
      if(nread != pktsize){
	cerr << "W @SUDPProto::RX_SUDP(): recvfrom() recevd(" << nread 
	     << "byte)/expected(" << pktsize << " byte)" << endl;
	pktsize = nread;
	framesize = pktsize - seqsize;
      }
      break;
    }
    seq = psudphdr->Seq();
    if(pktcnt ==0 || psudphdr->Cmd() == SudpHdr::INIT ){
      cerr << " Init-Packet Received"<< endl;
      seq0 = seq;
      if(txrx_length > 0) EndEpoch = time(0) + txrx_length;
    }
    w_pos = ( seq-seq0 ) * framesize;
    fseek(fp, w_pos, SEEK_SET);
    fwrite(frame_ptr,framesize,1,fp);
    total_wrote += framesize;
    //    *Logsystem::debug_logger() << *pseq << "(" << w_pos << ") ";
    
    rate_meas.Add_Rcv_Pld( framesize, seq );

    err_cnt = ERROR_CNT_LIMIT;
    pktcnt++;
    //    rcv_cnt++; // intvl_cnt++;
  }// while loop

  if(err_cnt <= 0){
    cerr << "E SUDPProt::RX_SUDP(): Leaving due to error " << endl;
    Timer_Stop();
    return 0;
  }
  if( psudphdr->Cmd() == SudpHdr::END ) cerr << " END packet received." << endl;
  if(EndEpoch >0 && time(0) > EndEpoch) cerr << " Recording time(" << txrx_length 
					     << "sec) end" << endl;
  Timer_Stop();
  return total_wrote;

}

void SUDPProto::Finish(){
  Timer_Stop();
  EndEpoch=0;
}

// /** RXvdif()
//  * @param arr[]: buffer to store the data.  
//  * @param pagesize: pagesize of the buffer
//  * @param npage: Number of buffer page
//  */
//const int MinPacketCount = 200;
//int RXvdif(vector <uint8_t>* arr, const int32_t pagesize, const int npage){
/** Receiving VDIF data stream 
 * @param dpage : Data_Page class to store the data
 */
int SUDPProto::RXvdif(DataPage& dpage){
  if(vlbi_frame->Type() != VLBI_Frame::VDIF_frm ){
    cerr << "E SUDPProt::RXvdif(): VDIF-frame was expected but, " 
	 << vlbi_frame->TypeName() << endl;
    return -1;
  }
  
  int mtu = _sock->get_mtu();
  int seqsize = sizeof(SudpHdr);  
  int headsize = vlbi_frame->Headsize();
  applProtocolState = RXING;
  //    status = RCVNG;
  //  vector <uint8_t> packet(mtu);
  packet.resize(mtu);
  //  const int64_t* pseq = (int64_t*)&packet[0];
  SudpHdr* psudphdr = (SudpHdr*)&packet[0];
  uint8_t* frame_ptr = &packet[0] + seqsize;
  uint8_t* dat_ptr = frame_ptr + headsize;
  memset(&packet[0],0,packet.size());

  int addr_len = sizeof(addr);
  //  runflag = true;
  applProtocolState = RXING;
  Timer_Start();
  int pktsize = mtu;
  int framesize = pktsize - seqsize;
  int pld_size = framesize - headsize;
  VDIF_Hdr* vdif_hdr_ptr = (VDIF_Hdr*)frame_ptr;
  int32_t speed = -1;
  int err_cnt= ERROR_CNT_LIMIT;
  int nread;

  /*  Inital reading data and Re-Configure the dpage/das_inf parametes. */
  struct timeval tm1, tm2;
  gettimeofday(&tm1,0);

  while( speed < 0 && err_cnt >0){
    nread = _sock->recvfrom((struct sockaddr&)addr, addr_len, 
			    (char*)&packet[0], pktsize);
    if( nread == 0 ) continue;
    if( nread < 0 ){
      cerr << "E SUDPProto::RX_SUDP():recvfrom() return:" 
	   << strerror(errno) << endl;
      err_cnt--; continue;
    }

    if( nread != pktsize ){
      cerr << "W @SUDPProto::RXvdif(): recevd packet(" << nread 
	   << "byte)/expected(" << pktsize << " byte)" << endl;
      pktsize = nread;
      if( framesize != pktsize - seqsize ){
	framesize = pktsize - seqsize;
	pld_size = framesize - headsize;
	if(vlbi_frame->Framesize() != framesize){
	  cerr << "W @SUPProto::RXvdif(): Received Framesize("
	       << framesize  << ") differ from " << vlbi_frame->TypeName()
	       << "->Framesize(" << vlbi_frame->Framesize() << ") Adjusted."
	       << endl;
	  Logsystem::info_logger()->Tstamp() 
	    << "W @SUPProto::RXvdif(): Received Framesize("
	    << framesize  << ") differ from " << vlbi_frame->TypeName()
	    << "->Framesize(" << vlbi_frame->Framesize() << ") Adjusted.\n";
	  vlbi_frame->Set_Framesize(framesize);
	}
      }
    } // end of if( nread != pktsize )

    gettimeofday(&tm2,0);
    if(tm2.tv_sec > tm1.tv_sec){ // Print Header every second. by M.S. 2013.6.15
      vdif_hdr_ptr->Print_hd(stderr);
      tm1 = tm2;
    }

    string stname,id;
    vdif_hdr_ptr->exud_nict.Get_Stname(stname);
    vdif_hdr_ptr->Get_Stnid(id);
    vlbi_frame->Set_StnameID(stname.c_str(),id.c_str());
    speed = vdif_hdr_ptr->Speed();
    if(speed > 0) {
      DAS_Info das_inf = vlbi_frame->DAS_Inf();
      if( das_inf.Nch() != vdif_hdr_ptr->Nch() ||
	  das_inf.Qbit() != vdif_hdr_ptr->Qbit() ||
	  das_inf.Srate() != (int32_t)vdif_hdr_ptr->Srate()){
	cerr << "W @SUPProto::RXvdif():Samping mode differs from expected(" 
	     << das_inf.Mode() 
	     << ") but recvd=" <<  vdif_hdr_ptr->Srate()/1000000 << "Mbps/"
	     <<  vdif_hdr_ptr->Qbit() << "bit/" <<  vdif_hdr_ptr->Nch() 
	     << "ch). DAS_Info updated." << endl;
	Logsystem::info_logger()->Tstamp() 		
	  << "W @SUPProto::RXvdif():Samping mode differs from expected(" 
	  << das_inf.Mode() 
	  << ") but recvd=" <<  vdif_hdr_ptr->Srate()/1000000 << "Mbps/"
	  <<  vdif_hdr_ptr->Qbit() << "bit/" <<  vdif_hdr_ptr->Nch() 
	  << "ch). DAS_Info updated.\n";
	das_inf.Set_Param( vdif_hdr_ptr->Nch(), vdif_hdr_ptr->Qbit(),
			   vdif_hdr_ptr->Srate());
	vlbi_frame->Set_DasInfo(das_inf);
      }
      int32_t pagesize = speed * 4;
      if( (int32_t)dpage.Pagesize() != pagesize ){
	cerr << "W @SUPProto::RXvdif():Pagesize Adjusted(" << pagesize
	     << ")" << endl;
	dpage.PageResize(pagesize);  
	speed=-1; // To get latest packet once again
      }
    } // if (speed >0)
  } // While loop 

  Time_dat hdrEpoch= vdif_hdr_ptr->Epoch(); 
  time_t epoch = hdrEpoch.Time_t();
  time_t epoch0 = epoch; // Initial epoch
  int ipage = epoch % dpage.Npage();
  int32_t refEpoch = vdif_hdr_ptr->Ref_Epoch().Time_t();
  //  int ipage_prev = ipage;
  Logsystem::debug_logger()->Tstamp() 
    << "RXvdif(): Rcving Page[" << ipage << "]" 
    << hdrEpoch.Print_time(10)   << "\n";
  Logsystem::debug_logger()->Flush();
  cerr  << "RXvdif(): Rcving Page[" << ipage << "]" 
	<< hdrEpoch.Print_time(10)   << "\n";

  while(dpage.IpageWriting() != ipage && dpage.Normal() ) {
    Logsystem::info_logger()->Tstamp()
      << "I SUDPProto::Rxvdif(): Filling EMPTY at " << dpage.IpageWriting() 
      << "pg unitl " << ipage << "pg\n";
    dpage.WStepPage(DataPage::EMPTY);
  }
  dpage.SetTimeWritingPage( epoch );
  int next_page = ( ipage +1 ) % dpage.Npage();
  int prev_page = ipage == 0 ? dpage.Npage() -1 : ipage -1;
  int32_t iframe = vdif_hdr_ptr->iFrame();
  int32_t ipos =  iframe * pld_size;

  if( ipos > (int32_t)dpage.Pagesize() ) {
    cerr << "E SUDPProto::RXvdif(): iframe(" << iframe
	 << ")*pld (" << ipos << ")is larger than pagesize("
	 << dpage.Pagesize() << ")" << endl; 
    applProtocolState = RX_READY;
    Timer_Stop();
    return -1;
  }
  //  uint8_t* buf_ptr = &dpage.dat[ipage][ipos];
  const uint8_t* buf_ptr = dpage.Data(ipage, ipos);
  memcpy( (void*)buf_ptr, dat_ptr, pld_size );
  rate_meas.Add_Rcv_Pld(pld_size, psudphdr->Seq());

  //  *Logsystem::debug_logger() << *pseq << " ";
  
  err_cnt = ERROR_CNT_LIMIT;
  while( err_cnt > 0 && applProtocolState == RXING ){
    int nread = _sock->recvfrom((struct sockaddr&)addr, addr_len,
				(char*)&packet[0], mtu);
    //    cerr << "Received pktsize=" << nread << endl;
    if(nread == 0){
      cerr << "E SUDPProto::RXvdif(): _sock->recvfrom() 0 pkt received" << endl;
      continue;
    }
    if(nread != pktsize){
      cerr << "E SUDPProto::RXvdif(): recvfrom() recevd(" << nread 
	   << "byte)/expected(" << pktsize << " byte)" << endl;
      Logsystem::error_logger()->Tstamp()
	<< "E SUDPProto::RXvdif(): recvfrom() recevd(" << nread 
	<< "byte)/expected(" << pktsize << " byte)\n";
      err_cnt--;
      continue;
    }

    //    epoch = vdif_hdr_ptr->Epoch().Time_t();
    epoch = vdif_hdr_ptr->Sec_Ref() + refEpoch;
    ipage = epoch % dpage.Npage();
    
    if( ipage != dpage.IpageWriting() ){
      hdrEpoch = vdif_hdr_ptr->Epoch();
      Logsystem::debug_logger()->Tstamp() 
	<< "RXvdif(): Rcving Page[" << ipage << "]" 
	<< hdrEpoch.Print_time(10)	<< "\n";
      Logsystem::debug_logger()->Flush();
      cerr << "RXvdif(): Rcving Page[" << ipage << "]" 
	<< hdrEpoch.Print_time(10)	<< endl;
      
      if(ipage == next_page ){ // next page
	prev_page = dpage.IpageWriting();
	dpage.WStepPage( DataPage::RD_OK, false ); // Prev page is released
	next_page = (dpage.IpageWriting() +1 ) % dpage.Npage();
	dpage.SetTimeWritingPage(epoch);
      }else
	if(ipage == prev_page){
	  cerr << " Recovering packet for previous page.\n";
	}else{
	  Logsystem::error_logger()->Tstamp()
	    << "E SUDPProt::RXvdif(): ipage(" << ipage
	     << ") Jump from current page("
	    << dpage.IpageWriting() << ")\n";
	  cerr << "E SUDPProt::RXvdif(): ipage(" << ipage
	       << ") Jump from current page("
	       << dpage.IpageWriting() << ")" << endl;
	  while(dpage.IpageWriting() != ipage && dpage.Normal()){
	    Logsystem::error_logger()->Tstamp()
	      << "E SUDPProto::Rxvdif(): Filling EMPTY at " 
	      << dpage.IpageWriting()
	      << "pg until " << ipage << "pg\n";
	    dpage.WStepPage(DataPage::EMPTY);
	  }
	  prev_page = ipage == 0 ? dpage.Npage() -1 : ipage -1;
	  next_page = (dpage.IpageWriting() +1 ) % dpage.Npage();
	  err_cnt--;
	  continue;
	}
    } // if (ipage != ...)
    uint32_t seq = psudphdr->Seq();
    rate_meas.Add_Rcv_Pld(pld_size, seq);

    ipos =  vdif_hdr_ptr->iFrame() * pld_size;
    //    *Logsystem::debug_logger() << ipos << " ";

    if( ( ipos+pld_size > (int32_t)dpage.Pagesize()) ){
      Logsystem::error_logger()->Tstamp() 
	<< "E SUDPProto::RXvdif(): ipos(" << ipos << ")>pagesize(" 
	<< dpage.Pagesize() << ") or #seq("<< seq << ") is out of range\n";
      err_cnt--;
      continue;
    }
    if(txrx_length > 0 &&  epoch > epoch0 + txrx_length){ 
      Time_dat t1(epoch0),t2(epoch);
      Logsystem::info_logger()->Tstamp()
	<< "RXvdif(): recording length(" << txrx_length 
	<< "sec) has passed for"
	<< t1.Print_time(10) << "-" << t2.Print_time(10) << "\n";
      //   dpage.ChgPageState(ipage,DataPage::DATA_END);
      cerr << "RXvdif(): recording length(" << txrx_length 
	<< "sec) has passed for"
	<< t1.Print_time(10) << "-" << t2.Print_time(10) << "\n";
      applProtocolState = RX_READY;
      //      runflag = false;
      continue;
    }

    //    buf_ptr = &(dpage.dat[ipage][ipos]);
    buf_ptr = dpage.Data(ipage,ipos);
    memcpy( (void*)buf_ptr, dat_ptr, pld_size);

    err_cnt = ERROR_CNT_LIMIT;

  }// while loop
  cerr << " applprotocolState=" << applProtocolState << endl;
  dpage.ChgPageState(ipage,DataPage::DATA_END);

  if(err_cnt <= 0){
    cerr << "E SUDPProt::RXvdif(): Leaving due to error " << endl;
    applProtocolState = RX_READY;
    Timer_Stop();
    return -1;
  }

  applProtocolState = RX_READY;
  Timer_Stop();
  return 0;
  
}

void SUDPProto::FirstPacket(){

  if(transLayerProtocol != UDP_IO){
    cerr << "E SUDPProto::FirstPacket(): SUDPProto is not tx by UDP\n";
    return ;
  }
  int pktsize = vlbi_frame->Framesize() + sizeof(SudpHdr);
  sudphdr.ResetSeq();
  sudphdr.SetFramesize(pktsize);
  sudphdr.SetValid();
  memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
  int loop = 3;
  int nwrote=0;
  my_usleep(16.0);
  while( loop > 0 && nwrote != pktsize){
    nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
    my_usleep(16.0);
    if(nwrote < 0 ) loop --;
  }

}

void SUDPProto::LastPacket(){

  if(transLayerProtocol != UDP_IO){
    cerr << "E SUDPProto::LastPacket(): SUDPProto is not tx by UDP\n";
    return ;
  }
  int pktsize = vlbi_frame->Framesize() + sizeof(SudpHdr);
  sudphdr.SetFramesize(pktsize);
  sudphdr.SetEND();
  memcpy(&packet[0],&sudphdr,sizeof(sudphdr));
  int loop = 3;
  my_usleep(16.0);
  int nwrote(0);
  while( loop > 0 && nwrote != pktsize){
    nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			   (char*)&packet[0], pktsize);
    my_usleep(16.0);
    if(nwrote < 0) loop --;
  }

}
// void SUDPProto::Set_VlbiFrame(VLBI_Frame * vf){
//   vlbi_frame = vf;
//   mtu = vf->Framesize();
// }

